# Copyright 2016 The Gemmlowp Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generates the arm32 headers used by the gemm/gemv lib."""

import cc_emitter
import common
import neon_emitter_64
import quantized_mul_kernels_common


def Main():
  """."""
  cc = cc_emitter.CCEmitter()
  common.GenerateHeader(cc, 'gemmlowp_meta_quantized_mul_kernels_arm_64',
                        'GEMMLOWP_NEON_64')

  cc.EmitNamespaceBegin('gemmlowp')
  cc.EmitNamespaceBegin('meta')
  cc.EmitNewline()

  shapes = [(1, 1), (1, 2), (1, 3), (1, 4), (1, 5), (1, 6), (1, 7), (1, 8),
            (2, 1), (2, 2), (2, 3), (2, 4), (3, 1), (3, 2), (3, 3)]

  quantized_mul_kernels_common.GenerateKernels(cc,
                                               neon_emitter_64.NeonEmitter64(),
                                               shapes)

  cc.EmitNamespaceEnd()
  cc.EmitNamespaceEnd()
  cc.EmitNewline()

  common.GenerateFooter(cc, 'Meta gemm for arm64 requires: GEMMLOWP_NEON_64!')


if __name__ == '__main__':
  Main()



# Copyright 2016 The Gemmlowp Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generates the arm32 headers used by the gemm/gemv lib."""

import cc_emitter
import common
import neon_emitter
import transform_kernels_common


def Main():
  """."""
  cc = cc_emitter.CCEmitter()
  common.GenerateHeader(cc, 'gemmlowp_meta_transform_kernels_arm_32',
                        'GEMMLOWP_NEON_32')

  cc.EmitNamespaceBegin('gemmlowp')
  cc.EmitNamespaceBegin('meta')
  cc.EmitNewline()

  transform_kernels_common.GenerateKernels(cc,
                                           neon_emitter.NeonEmitter(),
                                           [(16, x) for x in range(16)])

  cc.EmitNamespaceEnd()
  cc.EmitNamespaceEnd()
  cc.EmitNewline()

  common.GenerateFooter(cc, 'Meta gemm for arm32 requires: GEMMLOWP_NEON_32!')


if __name__ == '__main__':
  Main()



# Copyright 2016 The Gemmlowp Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""."""
import collections

_HEADER_COPYRIGHT = (
    '''// Copyright 2016 The Gemmlowp Authors. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
''')


def GenerateHeader(cc, header_name, preprocessor_directive):
  cc.EmitCodeNoSemicolon(_HEADER_COPYRIGHT)
  cc.EmitHeaderBegin(header_name)

  cc.EmitPreprocessor1('ifdef', preprocessor_directive)
  cc.EmitNewline()

  cc.EmitInclude('<cassert>')
  cc.EmitInclude('<cstdint>')
  cc.EmitNewline()


def GenerateFooter(cc, message):
  cc.EmitPreprocessor('else')
  cc.EmitPreprocessor1('warning', '"%s"' % message)
  cc.EmitPreprocessor('endif')
  cc.EmitNewline()
  cc.EmitHeaderEnd()


def GenerateDebugLog(cc, message):
  cc.EmitPreprocessor1('ifdef', 'DEBUG')
  cc.EmitPreprocessor1('ifdef', 'DEBUG_METAGEMM_VERBOSE')
  cc.EmitCode('std::cout << __FILE__ << \"(\" << __LINE__ << \") %s\" '
              '<< std::endl << std::flush' % message)
  cc.EmitPreprocessor('endif')
  cc.EmitPreprocessor('endif')


def _TemplateName(base, params):
  return '%s<%s>' % (base, ', '.join(map(str, params)))


class StreamGenerator(object):
  """."""

  def __init__(self, emitter, name):
    self.name = name
    self.emitter = emitter

  def SpecializeStream(self, in_type, lanes_count, pack_size, leftovers):
    if isinstance(getattr(self, 'EmitPack', None), collections.Callable):
      template_params = [in_type, lanes_count, pack_size, leftovers, self.name]
      self.emitter.EmitMemberFunctionBegin(
          'Stream', [], template_params, 'Pack',
          [['const %s*' % in_type, 'in'], ['const %s&' % self.name, 'params'],
           ['%s*' % in_type, 'out']], 'inline void')
      GenerateDebugLog(self.emitter,
                       '%s::Pack()' % _TemplateName(self.name, template_params))
      self.EmitPack(in_type, lanes_count, pack_size, leftovers)
      self.emitter.EmitFunctionEnd()


class MulKernelGenerator(object):
  """."""

  def __init__(self, emitter, kernel_name, output_stream_name):
    self.kernel_name = kernel_name
    self.output_stream_name = output_stream_name
    self.emitter = emitter

  def SpecializeMulKernel(self, in_type, out_type, kernel_m, kernel_n,
                          pack_size):
    """Generates the kernel wrapped in a MulKernel template specialization."""
    template_params = [
        in_type, out_type, self.kernel_name, self.output_stream_name, kernel_m,
        kernel_n, pack_size
    ]
    self.emitter.EmitMemberFunctionBegin(
        'MulKernel', [], template_params, 'Multiply',
        [['const %s*' % in_type, 'lhs'], ['const %s*' % in_type, 'rhs'], [
            'const FusedKernelParams<%s, %s>&' % (self.kernel_name,
                                                  self.output_stream_name),
            'params'
        ], ['%s*' % out_type, 'result']], 'inline void')
    GenerateDebugLog(self.emitter, '%s::Multiply()' %
                     _TemplateName(self.kernel_name + self.output_stream_name,
                                   template_params))
    self.EmitMultiply(in_type, out_type, kernel_m, kernel_n, pack_size)
    self.emitter.EmitFunctionEnd()


class Transform1DKernelGenerator(object):
  """."""

  def __init__(self, emitter, kernel_name):
    self.kernel_name = kernel_name
    self.emitter = emitter

  def SpecializeTransform1DKernel(self, in_type, out_type, kernel_size,
                                  leftovers):
    """Generates the kernel wrapped in a Transform1DKernel specialization."""
    template_params = [
        in_type, out_type, self.kernel_name, kernel_size, leftovers
    ]
    self.emitter.EmitMemberFunctionBegin(
        'Transform1DKernel', [], template_params, 'Transform',
        [['const %s*' % in_type, 'input'],
         ['const %s&' % self.kernel_name, 'params'],
         ['%s*' % out_type, 'output']], 'inline void')
    GenerateDebugLog(self.emitter, '%s::Transform()' %
                     _TemplateName(self.kernel_name, template_params))
    self.EmitTransform(in_type, out_type, kernel_size, leftovers)
    self.emitter.EmitFunctionEnd()



# Copyright 2016 The Gemmlowp Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""32bit ARM/NEON assembly emitter.

Used by code generators to produce ARM assembly with NEON simd code.
Provides tools for easier register management: named register variable
allocation/deallocation, and offers a more procedural/structured approach
to generating assembly.

TODO: right now neon emitter prints out assembly instructions immediately,
it might be beneficial to keep the whole structure and emit the assembly after
applying some optimizations like: instruction reordering or register reuse.

TODO: NeonRegister object assigns explicit registers at allocation time.
Similarily to emiting code, register mapping and reuse can be performed and
optimized lazily.
"""


class Error(Exception):
  """Module level error."""


class RegisterAllocationError(Error):
  """Cannot alocate registers."""


class LaneError(Error):
  """Wrong lane number."""


class ArgumentError(Error):
  """Wrong argument."""


def _Low(register):
  assert register[0] == 'q'
  num = int(register[1:])
  return 'd%d' % (num * 2)


def _High(register):
  assert register[0] == 'q'
  num = int(register[1:])
  return 'd%d' % (num * 2 + 1)


def _ExpandQuads(registers):
  doubles = []
  for register in registers:
    if register[0] == 'q':
      doubles.append(_Low(register))
      doubles.append(_High(register))
    else:
      doubles.append(register)
  return doubles


def _MakeCompatible(op1, op2, op3):
  if op1[0] == 'd' or op2[0] == 'd' or op3[0] == 'd':
    if op1[0] == 'q':
      op1 = _Low(op1)
    if op2[0] == 'q':
      op2 = _Low(op2)
    if op3[0] == 'q':
      op3 = _Low(op3)
  return (op1, op2, op3)


class _NeonRegisters32Bit(object):
  """Utility that keeps track of used 32bit ARM/NEON registers."""

  def __init__(self):
    self.double = set()
    self.double_ever = set()
    self.general = set()
    self.general_ever = set()
    self.parameters = dict()
    self.output_parameters = dict()

  def MapParameter(self, parameter, parameter_value=None):
    if not parameter_value:
      parameter_value = parameter
    self.parameters[parameter] = (parameter_value, 'r')
    return '%%[%s]' % parameter

  def MapMemoryParameter(self, parameter, parameter_value=None):
    if not parameter_value:
      parameter_value = parameter
    self.parameters[parameter] = (parameter_value, 'm')
    return '%%[%s]' % parameter

  def MapOutputParameter(self, parameter, parameter_value=None):
    if not parameter_value:
      parameter_value = parameter
    self.output_parameters[parameter] = (parameter_value, '+r')
    return '%%[%s]' % parameter

  def DoubleRegister(self, min_val=0):
    for i in range(min_val, 32):
      if i not in self.double:
        self.double.add(i)
        self.double_ever.add(i)
        return 'd%d' % i
    raise RegisterAllocationError('Not enough double registers.')

  def QuadRegister(self, min_val=0):
    for i in range(min_val, 16):
      if ((i * 2) not in self.double) and ((i * 2 + 1) not in self.double):
        self.double.add(i * 2)
        self.double.add(i * 2 + 1)
        self.double_ever.add(i * 2)
        self.double_ever.add(i * 2 + 1)
        return 'q%d' % i
    raise RegisterAllocationError('Not enough quad registers.')

  def GeneralRegister(self):
    for i in range(0, 16):
      if i not in self.general:
        self.general.add(i)
        self.general_ever.add(i)
        return 'r%d' % i
    raise RegisterAllocationError('Not enough general registers.')

  def MappedParameters(self):
    return [(k, v) for (k, v) in self.parameters.items()]

  def MappedOutputParameters(self):
    return [(k, v) for (k, v) in self.output_parameters.items()]

  def Clobbers(self):
    return (['r%d' % i for i in self.general_ever] +
            ['d%d' % i for i in self.DoubleClobbers()])

  def DoubleClobbers(self):
    return sorted(self.double_ever)

  def FreeRegister(self, register):
    assert len(register) > 1
    if register[0] not in ['r', 'd', 'q']:
      return

    num = int(register[1:])

    if register[0] == 'r':
      assert num in self.general
      self.general.remove(num)
    elif register[0] == 'd':
      assert num in self.double
      self.double.remove(num)
    elif register[0] == 'q':
      assert num * 2 in self.double
      assert num * 2 + 1 in self.double
      self.double.remove(num * 2)
      self.double.remove(num * 2 + 1)
    else:
      raise RegisterDeallocationError('Register not allocated: %s' % register)

  def FreeRegisters(self, registers):
    for register in registers:
      self.FreeRegister(register)


class NeonEmitter(object):
  """Emits ARM/NEON assembly opcodes."""

  def __init__(self, debug=False):
    self.ops = {}
    self.indent = ''
    self.debug = debug

  def PushIndent(self, delta='  '):
    self.indent += delta

  def PopIndent(self, delta=2):
    self.indent = self.indent[:-delta]

  def EmitIndented(self, what):
    print(self.indent + what)

  def PushOp(self, op):
    if op in self.ops.keys():
      self.ops[op] += 1
    else:
      self.ops[op] = 1

  def ClearCounters(self):
    self.ops.clear()

  def EmitNewline(self):
    print('')

  def EmitPreprocessor1(self, op, param):
    print('#%s %s' % (op, param))

  def EmitPreprocessor(self, op):
    print('#%s' % op)

  def EmitInclude(self, include):
    self.EmitPreprocessor1('include', include)

  def EmitCall1(self, function, param):
    self.EmitIndented('%s(%s);' % (function, param))

  def EmitAssert(self, assert_expression):
    if self.debug:
      self.EmitCall1('assert', assert_expression)

  def EmitHeaderBegin(self, header_name, includes):
    self.EmitPreprocessor1('ifndef', (header_name + '_H_').upper())
    self.EmitPreprocessor1('define', (header_name + '_H_').upper())
    self.EmitNewline()
    if includes:
      for include in includes:
        self.EmitInclude(include)
      self.EmitNewline()

  def EmitHeaderEnd(self):
    self.EmitPreprocessor('endif')

  def EmitCode(self, code):
    self.EmitIndented('%s;' % code)

  def EmitFunctionBeginA(self, function_name, params, return_type):
    self.EmitIndented('%s %s(%s) {' %
                      (return_type, function_name,
                       ', '.join(['%s %s' % (t, n) for (t, n) in params])))
    self.PushIndent()

  def EmitFunctionEnd(self):
    self.PopIndent()
    self.EmitIndented('}')

  def EmitAsmBegin(self):
    self.EmitIndented('asm volatile(')
    self.PushIndent()

  def EmitAsmMapping(self, elements):
    if elements:
      self.EmitIndented(': ' + ', '.join(
          ['[%s] "%s"(%s)' % (d, v[1], v[0]) for (d, v) in elements]))
    else:
      self.EmitIndented(':')

  def EmitClobbers(self, elements):
    if elements:
      self.EmitIndented(': ' + ', '.join(['"%s"' % c for c in elements]))
    else:
      self.EmitIndented(':')

  def EmitAsmEnd(self, registers):
    self.EmitAsmMapping(registers.MappedOutputParameters())
    self.EmitAsmMapping(registers.MappedParameters())
    self.EmitClobbers(registers.Clobbers() + ['cc', 'memory'])
    self.PopIndent()
    self.EmitIndented(');')

  def EmitComment(self, comment):
    self.EmitIndented('// ' + comment)

  def EmitNumericalLabel(self, label):
    self.EmitIndented('"%d:"' % label)

  def EmitOp1(self, op, param1):
    self.PushOp(op)
    self.EmitIndented('"%s %s\\n"' % (op, param1))

  def EmitOp2(self, op, param1, param2):
    self.PushOp(op)
    self.EmitIndented('"%s %s, %s\\n"' % (op, param1, param2))

  def EmitOp3(self, op, param1, param2, param3):
    self.PushOp(op)
    self.EmitIndented('"%s %s, %s, %s\\n"' % (op, param1, param2, param3))

  def EmitAdd(self, destination, source, param):
    self.EmitOp3('add', destination, source, param)

  def EmitSubs(self, destination, source, param):
    self.EmitOp3('subs', destination, source, param)

  def EmitSub(self, destination, source, param):
    self.EmitOp3('sub', destination, source, param)

  def EmitMul(self, destination, source, param):
    self.EmitOp3('mul', destination, source, param)

  def EmitMov(self, param1, param2):
    self.EmitOp2('mov', param1, param2)

  def EmitBeqBack(self, label):
    self.EmitOp1('beq', '%db' % label)

  def EmitBeqFront(self, label):
    self.EmitOp1('beq', '%df' % label)

  def EmitBgtBack(self, label):
    self.EmitOp1('bgt', '%db' % label)

  def EmitBgtFront(self, label):
    self.EmitOp1('bgt', '%df' % label)

  def EmitBleBack(self, label):
    self.EmitOp1('ble', '%db' % label)

  def EmitBleFront(self, label):
    self.EmitOp1('ble', '%df' % label)

  def EmitBneBack(self, label):
    self.EmitOp1('bne', '%db' % label)

  def EmitBneFront(self, label):
    self.EmitOp1('bne', '%df' % label)

  def EmitVAdd(self, add_type, destination, source_1, source_2):
    destination, source_1, source_2 = _MakeCompatible(destination, source_1,
                                                      source_2)
    self.EmitOp3('vadd.%s' % add_type, destination, source_1, source_2)

  def EmitVAddw(self, add_type, destination, source_1, source_2):
    self.EmitOp3('vaddw.%s' % add_type, destination, source_1, source_2)

  def EmitVSub(self, sub_type, destination, source_1, source_2):
    destination, source_1, source_2 = _MakeCompatible(destination, source_1,
                                                      source_2)
    self.EmitOp3('vsub.%s' % sub_type, destination, source_1, source_2)

  def EmitVCvt(self, cvt_to, cvt_from, destination, source):
    self.EmitOp2('vcvt.%s.%s' % (cvt_to, cvt_from), destination, source)

  def EmitVDup(self, dup_type, destination, source):
    self.EmitOp2('vdup.%s' % dup_type, destination, source)

  def EmitVMax(self, size, destination, source_1, source_2):
    self.EmitOp3('vmax.%s' % size, destination, source_1, source_2)

  def EmitVMin(self, size, destination, source_1, source_2):
    self.EmitOp3('vmin.%s' % size, destination, source_1, source_2)

  def EmitVMov(self, mov_type, destination, source):
    self.EmitOp2('vmov.%s' % mov_type, destination, source)

  def EmitVMovl(self, mov_type, destination, source):
    if source[0] == 'q':
      source = _Low(source)
    self.EmitOp2('vmovl.%s' % mov_type, destination, source)

  def EmitVMovl2(self, mov_type, destination_1, destination_2, source):
    self.EmitVMovl(mov_type, destination_2, _High(source))
    self.EmitVMovl(mov_type, destination_1, _Low(source))

  def EmitVQmovn(self, mov_type, destination, source):
    if destination[0] == 'q':
      destination = _Low(destination)
    self.EmitOp2('vqmovn.%s' % mov_type, destination, source)

  def EmitVQmovn2(self, mov_type, destination, source_1, source_2):
    self.EmitVQmovn(mov_type, _Low(destination), source_1)
    self.EmitVQmovn(mov_type, _High(destination), source_2)

  def EmitVQmovun(self, mov_type, destination, source):
    if destination[0] == 'q':
      destination = _Low(destination)
    self.EmitOp2('vqmovun.%s' % mov_type, destination, source)

  def EmitVQmovun2(self, mov_type, destination, source_1, source_2):
    self.EmitVQmovun(mov_type, _Low(destination), source_1)
    self.EmitVQmovun(mov_type, _High(destination), source_2)

  def EmitVMul(self, mul_type, destination, source_1, source_2):
    destination, source_1, source_2 = _MakeCompatible(destination, source_1,
                                                      source_2)
    self.EmitOp3('vmul.%s' % mul_type, destination, source_1, source_2)

  def EmitVMulScalar(self, mul_type, destination, source_1, source_2):
    self.EmitOp3('vmul.%s' % mul_type, destination, source_1, source_2)

  def EmitVMull(self, mul_type, destination, source_1, source_2):
    self.EmitOp3('vmull.%s' % mul_type, destination, source_1, source_2)

  def EmitVPadd(self, add_type, destination, source_1, source_2):
    self.EmitOp3('vpadd.%s' % add_type, destination, source_1, source_2)

  def EmitVPaddl(self, add_type, destination, source):
    self.EmitOp2('vpaddl.%s' % add_type, destination, source)

  def EmitVPadal(self, add_type, destination, source):
    self.EmitOp2('vpadal.%s' % add_type, destination, source)

  def EmitLdr(self, register, value):
    self.EmitOp2('ldr', register, value)

  def EmitVLoad(self, load_no, load_type, destination, source):
    self.EmitVLoadA(load_no, load_type, [destination], source)

  def EmitVLoadA(self, load_no, load_type, destinations, source):
    self.EmitOp2('vld%d.%d' % (load_no, load_type),
                 '{%s}' % ', '.join(_ExpandQuads(destinations)), source)

  def EmitVLoadAE(self,
                  load_type,
                  elem_count,
                  destinations,
                  source,
                  alignment=None):
    bits_to_load = load_type * elem_count
    destinations = _ExpandQuads(destinations)
    if len(destinations) * 64 < bits_to_load:
      raise ArgumentError('To few destinations: %d to load %d bits.' %
                          (len(destinations), bits_to_load))

    while bits_to_load > 0:
      if bits_to_load >= 256:
        self.EmitVLoadA(1, 32, destinations[:4],
                        self.DereferenceIncrement(source, alignment))
        bits_to_load -= 256
        destinations = destinations[4:]
      elif bits_to_load >= 192:
        self.EmitVLoadA(1, 32, destinations[:3],
                        self.DereferenceIncrement(source, alignment))
        bits_to_load -= 192
        destinations = destinations[3:]
      elif bits_to_load >= 128:
        self.EmitVLoadA(1, 32, destinations[:2],
                        self.DereferenceIncrement(source, alignment))
        bits_to_load -= 128
        destinations = destinations[2:]
      elif bits_to_load >= 64:
        self.EmitVLoad(1, 32, destinations[0],
                       self.DereferenceIncrement(source, alignment))
        bits_to_load -= 64
        destinations = destinations[1:]
      else:
        destination = destinations[0]
        if bits_to_load == 56:
          self.EmitVLoad(1, 32,
                         self.Lane(32, destination, 0),
                         self.DereferenceIncrement(source))
          self.EmitVLoad(1, 16,
                         self.Lane(16, destination, 2),
                         self.DereferenceIncrement(source))
          self.EmitVLoad(1, 8,
                         self.Lane(8, destination, 6),
                         self.DereferenceIncrement(source))
        elif bits_to_load == 48:
          self.EmitVLoad(1, 32,
                         self.Lane(32, destination, 0),
                         self.DereferenceIncrement(source))
          self.EmitVLoad(1, 16,
                         self.Lane(16, destination, 2),
                         self.DereferenceIncrement(source))
        elif bits_to_load == 40:
          self.EmitVLoad(1, 32,
                         self.Lane(32, destination, 0),
                         self.DereferenceIncrement(source))
          self.EmitVLoad(1, 8,
                         self.Lane(8, destination, 4),
                         self.DereferenceIncrement(source))
        elif bits_to_load == 32:
          self.EmitVLoad(1, 32,
                         self.Lane(32, destination, 0),
                         self.DereferenceIncrement(source))
        elif bits_to_load == 24:
          self.EmitVLoad(1, 16,
                         self.Lane(16, destination, 0),
                         self.DereferenceIncrement(source))
          self.EmitVLoad(1, 8,
                         self.Lane(8, destination, 2),
                         self.DereferenceIncrement(source))
        elif bits_to_load == 16:
          self.EmitVLoad(1, 16,
                         self.Lane(16, destination, 0),
                         self.DereferenceIncrement(source))
        elif bits_to_load == 8:
          self.EmitVLoad(1, 8,
                         self.Lane(8, destination, 0),
                         self.DereferenceIncrement(source))
        else:
          raise ArgumentError('Wrong leftover: %d' % bits_to_load)
        return

  def EmitVLoadE(self, load_type, count, destination, source, alignment=None):
    self.EmitVLoadAE(load_type, count, [destination], source, alignment)

  def EmitVLoadAllLanes(self, load_no, load_type, destination, source):
    destinations = []
    if destination[0] == 'q':
      destinations.append(self.AllLanes(_Low(destination)))
      destinations.append(self.AllLanes(_High(destination)))
    else:
      destinations.append(self.AllLanes(destination))
    self.EmitVLoadA(load_no, load_type, destinations, source)

  def EmitVLoadOffset(self, load_no, load_type, destination, source, offset):
    self.EmitVLoadOffsetA(load_no, load_type, [destination], source, offset)

  def EmitVLoadOffsetA(self, load_no, load_type, destinations, source, offset):
    assert len(destinations) <= 4
    self.EmitOp3('vld%d.%d' % (load_no, load_type),
                 '{%s}' % ', '.join(_ExpandQuads(destinations)), source, offset)

  def EmitPld(self, load_address_register):
    self.EmitOp1('pld', '[%s]' % load_address_register)

  def EmitPldw(self, store_address_register):
    self.EmitOp1('pldw', '[%s]' % store_address_register)

  def EmitPldOffset(self, load_address_register, offset):
    self.EmitOp1('pld', '[%s, %s]' % (load_address_register, offset))

  def EmitPldwOffset(self, store_address_register, offset):
    self.EmitOp1('pldw', '[%s, %s]' % (store_address_register, offset))

  def EmitVShl(self, shift_type, destination, source, shift):
    self.EmitOp3('vshl.%s' % shift_type, destination, source, shift)

  def EmitVStore(self, store_no, store_type, source, destination):
    self.EmitVStoreA(store_no, store_type, [source], destination)

  def EmitVStoreA(self, store_no, store_type, sources, destination):
    self.EmitOp2('vst%d.%d' % (store_no, store_type),
                 '{%s}' % ', '.join(_ExpandQuads(sources)), destination)

  def EmitVStoreAE(self,
                   store_type,
                   elem_count,
                   sources,
                   destination,
                   alignment=None):
    bits_to_store = store_type * elem_count
    sources = _ExpandQuads(sources)
    if len(sources) * 64 < bits_to_store:
      raise ArgumentError('To few sources: %d to store %d bits.' %
                          (len(sources), bits_to_store))

    while bits_to_store > 0:
      if bits_to_store >= 256:
        self.EmitVStoreA(1, 32, sources[:4],
                         self.DereferenceIncrement(destination, alignment))
        bits_to_store -= 256
        sources = sources[4:]
      elif bits_to_store >= 192:
        self.EmitVStoreA(1, 32, sources[:3],
                         self.DereferenceIncrement(destination, alignment))
        bits_to_store -= 192
        sources = sources[3:]
      elif bits_to_store >= 128:
        self.EmitVStoreA(1, 32, sources[:2],
                         self.DereferenceIncrement(destination, alignment))
        bits_to_store -= 128
        sources = sources[2:]
      elif bits_to_store >= 64:
        self.EmitVStore(1, 32, sources[0],
                        self.DereferenceIncrement(destination, alignment))
        bits_to_store -= 64
        sources = sources[1:]
      else:
        source = sources[0]
        if bits_to_store == 56:
          self.EmitVStore(1, 32,
                          self.Lane(32, source, 0),
                          self.DereferenceIncrement(destination))
          self.EmitVStore(1, 16,
                          self.Lane(16, source, 2),
                          self.DereferenceIncrement(destination))
          self.EmitVStore(1, 8,
                          self.Lane(8, source, 6),
                          self.DereferenceIncrement(destination))
        elif bits_to_store == 48:
          self.EmitVStore(1, 32,
                          self.Lane(32, source, 0),
                          self.DereferenceIncrement(destination))
          self.EmitVStore(1, 16,
                          self.Lane(16, source, 2),
                          self.DereferenceIncrement(destination))
        elif bits_to_store == 40:
          self.EmitVStore(1, 32,
                          self.Lane(32, source, 0),
                          self.DereferenceIncrement(destination))
          self.EmitVStore(1, 8,
                          self.Lane(8, source, 4),
                          self.DereferenceIncrement(destination))
        elif bits_to_store == 32:
          self.EmitVStore(1, 32,
                          self.Lane(32, source, 0),
                          self.DereferenceIncrement(destination))
        elif bits_to_store == 24:
          self.EmitVStore(1, 16,
                          self.Lane(16, source, 0),
                          self.DereferenceIncrement(destination))
          self.EmitVStore(1, 8,
                          self.Lane(8, source, 2),
                          self.DereferenceIncrement(destination))
        elif bits_to_store == 16:
          self.EmitVStore(1, 16,
                          self.Lane(16, source, 0),
                          self.DereferenceIncrement(destination))
        elif bits_to_store == 8:
          self.EmitVStore(1, 8,
                          self.Lane(8, source, 0),
                          self.DereferenceIncrement(destination))
        else:
          raise ArgumentError('Wrong leftover: %d' % bits_to_store)
        return

  def EmitVStoreE(self, store_type, count, source, destination, alignment=None):
    self.EmitVStoreAE(store_type, count, [source], destination, alignment)

  def EmitVStoreOffset(self, store_no, store_type, source, destination, offset):
    self.EmitVStoreOffsetA(store_no, store_type, [source], destination, offset)

  def EmitVStoreOffsetA(self, store_no, store_type, sources, destination,
                        offset):
    self.EmitOp3('vst%d.%d' % (store_no, store_type),
                 '{%s}' % ', '.join(_ExpandQuads(sources)), destination, offset)

  def EmitVStoreOffsetE(self, store_type, count, source, destination, offset):
    """Emit assembly to store a number elements from the source registers."""
    if store_type is not 32:
      raise ArgumentError('Unsupported store_type: %d' % store_type)

    sources = []
    if source[0] == 'q':
      sources.append(_Low(source))
      sources.append(_High(source))
      if count * store_type > 128:
        raise ArgumentError('To many %dbit elements in a q register: %d' %
                            (store_type, count))
    else:
      sources.append(source)
      if count * store_type > 64:
        raise ArgumentError('To many %dbit elements in a d register: %d' %
                            (store_type, count))

    if count == 1:
      self.EmitVStoreOffset(1, store_type,
                            self.Lane(store_type, sources[0], 0),
                            self.Dereference(destination, None), offset)
    elif count == 2:
      self.EmitVStoreOffset(1, store_type, sources[0],
                            self.Dereference(destination, None), offset)
    elif count == 3:
      self.EmitVStore(1, store_type, sources[0],
                      self.DereferenceIncrement(destination, None))
      self.EmitVStoreOffset(1, store_type,
                            self.Lane(store_type, sources[1], 0),
                            self.Dereference(destination, None), offset)
      self.EmitSub(destination, destination, self.ImmediateConstant(8))
    elif count == 4:
      self.EmitVStoreOffsetA(1, store_type, sources,
                             self.Dereference(destination, None), offset)
    else:
      raise ArgumentError('To many elements: %d' % count)

  def EmitVSumReduce(self, reduce_type, elem_count, reduce_count, destinations,
                     sources):
    """Emit assembly for n-fold horizontal sum reduction."""
    if reduce_type is not 'u32':
      raise ArgumentError('Unsupported reduce: %s' % reduce_type)

    sources = _ExpandQuads(sources)

    destinations = _ExpandQuads(destinations)

    if len(destinations) * 2 < elem_count:
      raise ArgumentError('Not enough space in destination: %d vs %d' %
                          (len(destinations) * 2, elem_count))

    if len(sources) * 2 != elem_count * reduce_count:
      raise ArgumentError('Wrong number of sources: %d vs %d' %
                          (len(sources) * 2, elem_count * reduce_count))

    if reduce_count <= 1:
      raise ArgumentError('Unsupported reduce_count: %d' % reduce_count)

    while reduce_count > 1:
      if len(sources) % 2 == 1:
        sources.append(sources[-1])

      if reduce_count == 2:
        for i in range(len(sources) / 2):
          self.EmitVPadd(reduce_type, destinations[i], sources[2 * i],
                         sources[2 * i + 1])
        return
      else:
        sources_2 = []
        for i in range(len(sources) / 2):
          self.EmitVPadd(reduce_type, sources[2 * i], sources[2 * i],
                         sources[2 * i + 1])
          sources_2.append(sources[2 * i])
        reduce_count /= 2
        sources = sources_2

  def EmitVUzp(self, uzp_type, operand_1, operand_2):
    self.EmitOp2('vuzp.%d' % uzp_type, operand_1, operand_2)

  def EmitVTrn(self, trn_type, operand_1, operand_2):
    self.EmitOp2('vtrn.%d' % trn_type, operand_1, operand_2)

  def EmitColBlockStride(self, cols, stride, new_stride):
    assert cols in [1, 2, 3, 4, 5, 6, 7, 8]
    if cols in [5, 6, 7]:
      self.EmitSub(new_stride, stride, self.ImmediateConstant(4))

  def EmitLoadColBlock(self, unused_registers, load_type, cols, elements, block,
                       input_address, stride):
    """Load a block of column major data."""
    assert cols is len(block)
    assert load_type is 8

    input_deref = self.Dereference(input_address, None)
    input_deref_increment = self.DereferenceIncrement(input_address, None)

    if cols is 1:
      for i in range(elements):
        self.EmitVLoadOffset(1, 8,
                             self.Lane(8, block[0], i), input_deref, stride)
      self.EmitPld(input_address)
    elif cols is 2:
      for i in range(elements):
        self.EmitVLoadOffset(1, 16,
                             self.Lane(16, block[i / 4], i % 4), input_deref,
                             stride)
      self.EmitPld(input_address)
      self.EmitVUzp(8, block[0], block[1])
    elif cols is 3:
      for i in range(elements):
        self.EmitVLoadOffsetA(3, 8, [self.Lane(8, row, i) for row in block],
                              input_deref, stride)
    elif cols is 4:
      for i in range(elements):
        self.EmitVLoadOffset(1, 32,
                             self.Lane(32, block[i % 4], i / 4), input_deref,
                             stride)
      self.EmitPld(input_address)
      self.EmitVTrn(16, block[0], block[2])
      self.EmitVTrn(16, block[1], block[3])
      self.EmitVTrn(8, block[0], block[1])
      self.EmitVTrn(8, block[2], block[3])
    elif cols is 5:
      for i in range(elements):
        self.EmitVLoad(1, 32,
                       self.Lane(32, block[i % 4], i / 4),
                       input_deref_increment)
        self.EmitVLoadOffset(1, 8,
                             self.Lane(8, block[4], i), input_deref, stride)
      self.EmitPld(input_address)
      self.EmitVTrn(16, block[0], block[2])
      self.EmitVTrn(16, block[1], block[3])
      self.EmitVTrn(8, block[0], block[1])
      self.EmitVTrn(8, block[2], block[3])
    elif cols is 6:
      for i in range(elements):
        self.EmitVLoad(1, 32,
                       self.Lane(32, block[i % 4], i / 4),
                       input_deref_increment)
        self.EmitVLoadOffset(1, 16,
                             self.Lane(16, block[4 + i / 4], i % 4),
                             input_deref, stride)
      self.EmitPld(input_address)
      self.EmitVTrn(16, block[0], block[2])
      self.EmitVTrn(16, block[1], block[3])
      self.EmitVUzp(8, block[4], block[5])
      self.EmitVTrn(8, block[0], block[1])
      self.EmitVTrn(8, block[2], block[3])
    elif cols is 7:
      for i in range(elements):
        self.EmitVLoad(1, 32,
                       self.Lane(32, block[i % 4], i / 4),
                       input_deref_increment)
        self.EmitVLoadOffsetA(3, 8,
                              [self.Lane(8, row, i) for row in block[4:]],
                              input_deref, stride)
      self.EmitPld(input_address)
      self.EmitVTrn(16, block[0], block[2])
      self.EmitVTrn(16, block[1], block[3])
      self.EmitVTrn(8, block[0], block[1])
      self.EmitVTrn(8, block[2], block[3])
    elif cols is 8:
      for i in range(elements):
        self.EmitVLoadOffset(1, 32, block[i], input_deref, stride)
      self.EmitPld(input_address)
      self.EmitVTrn(8, block[0], block[1])
      self.EmitVTrn(8, block[2], block[3])
      self.EmitVTrn(8, block[4], block[5])
      self.EmitVTrn(8, block[6], block[7])
      self.EmitVTrn(16, block[0], block[2])
      self.EmitVTrn(16, block[1], block[3])
      self.EmitVTrn(16, block[4], block[6])
      self.EmitVTrn(16, block[5], block[7])
      self.EmitVTrn(32, block[0], block[4])
      self.EmitVTrn(32, block[1], block[5])
      self.EmitVTrn(32, block[2], block[6])
      self.EmitVTrn(32, block[3], block[7])
    else:
      assert False
    return block

  def Dereference(self, value, alignment=None):
    if alignment:
      return '[%s:%d]' % (value, alignment)
    else:
      return '[%s]' % value

  def DereferenceIncrement(self, value, alignment=None):
    return '%s!' % self.Dereference(value, alignment)

  def ImmediateConstant(self, value):
    return '#%d' % value

  def AllLanes(self, value):
    return '%s[]' % value

  def Lane(self, bits, value, lane):
    """Get the proper n-bit lane from the given register."""
    registers = []
    if value[0] == 'q':
      registers.append(_Low(value))
      registers.append(_High(value))
    else:
      registers.append(value)

    elems_per_register = 64 / bits
    register = lane / elems_per_register
    lane %= elems_per_register

    return '%s[%d]' % (registers[register], lane)

  def CreateRegisters(self):
    return _NeonRegisters32Bit()



# Copyright 2016 The Gemmlowp Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""."""

import common


def _ReadParams(emitter, registers, input_address, elements, min_register):
  registers_count = (elements + 3) / 4
  registers = [
      registers.QuadRegister(min_register)
      for unused_i in range(registers_count)
  ]
  emitter.EmitVLoadAE(registers_count * 4, 32, registers, input_address, 64)
  return registers


def _Duplicate(emitter, registers, rows, values):
  """Populate a grid of registers duplicating provided values."""
  duplicated = []
  for i in range(rows):
    if i is rows - 1:
      duplicated.append(values[0])
    else:
      duplicated.append(registers.QuadRegister())

    emitter.EmitVDup('32', duplicated[i],
                     emitter.Lane(32, values[i / 4], i % 4))

  return duplicated


def _DuplicateGeneralRegister(emitter, registers, value, min_register):
  register = registers.QuadRegister(min_register)
  emitter.EmitVDup('32', register, value)
  return register


class _StaticQuantizationUInt8Transformation(object):
  """Calculate quantized values and cast back to uint8."""

  def Prepare(self, emitter, registers, kernel_m, kernel_n, lhs, rhs):
    """Load parameters and prepare duplicated registers."""
    emitter.EmitNewline()
    emitter.EmitComment('StaticQuantization::Prepare')

    lhs_offset = _ReadParams(emitter, registers, lhs, kernel_m, 4)
    self.rhs_offsets = _ReadParams(emitter, registers, rhs, kernel_n, 4)
    self.multiplicative_offset = _DuplicateGeneralRegister(
        emitter, registers,
        registers.MapParameter('multiplicative_offset',
                               'params.kernel.multiplicative_offset'), 4)
    self.rounding_offset = _DuplicateGeneralRegister(
        emitter, registers,
        registers.MapParameter('rounding_offset',
                               'params.kernel.rounding_offset'), 4)
    self.shift = _DuplicateGeneralRegister(
        emitter, registers,
        registers.MapParameter('shift', 'params.kernel.shift'), 4)
    self.lhs_offsets = _Duplicate(emitter, registers, kernel_m, lhs_offset)

  def Transform(self, emitter, registers, data, unused_kernel_m,
                unused_kernel_n):
    """Quantize the data."""
    emitter.EmitNewline()
    emitter.EmitComment('StaticQuantization::Transform')

    for (row, lhs_offset) in zip(data, self.lhs_offsets):
      for row_register in row:
        emitter.EmitVAdd('s32', row_register, row_register, lhs_offset)

    for row in data:
      for (row_register, rhs_offset_register) in zip(row, self.rhs_offsets):
        emitter.EmitVAdd('s32', row_register, row_register, rhs_offset_register)

    for row in data:
      for row_register in row:
        emitter.EmitVMul('i32', row_register, row_register,
                         self.multiplicative_offset)

    for row in data:
      for row_register in row:
        emitter.EmitVAdd('i32', row_register, row_register,
                         self.rounding_offset)

    for row in data:
      for row_register in row:
        emitter.EmitVShl('s32', row_register, row_register, self.shift)

    if len(data[0]) is 1:
      for row in data:
        emitter.EmitVQmovn('s32', row[0], row[0])

      for row in data:
        emitter.EmitVQmovun('s16', row[0], row[0])

      return data
    elif len(data[0]) is 2:
      results = []
      for row in data:
        emitter.EmitVQmovn2('s32', row[0], row[0], row[1])
        registers.FreeRegister(row[1])
        results.append([row[0]])

      for row in results:
        emitter.EmitVQmovun('s16', row[0], row[0])

      return results
    else:
      assert False

  def Type(self):
    return 8


class _StaticQuantizationInt32Transformation(object):
  """."""

  def Prepare(self, emitter, registers, kernel_m, kernel_n, lhs, rhs):
    emitter.EmitNewline()
    emitter.EmitComment('StaticQuantizationInt32::Prepare')

    lhs_offset = _ReadParams(emitter, registers, lhs, kernel_m, 4)
    self.rhs_offsets = _ReadParams(emitter, registers, rhs, kernel_n, 4)
    self.lhs_offsets = _Duplicate(emitter, registers, kernel_m, lhs_offset)

  def Transform(self, emitter, unused_registers, data, unused_kernel_m,
                unused_kernel_n):
    """Quantize data and output as int32."""
    emitter.EmitNewline()
    emitter.EmitComment('StaticQuantizationInt32::Transform')

    for (row, lhs_offset) in zip(data, self.lhs_offsets):
      for row_register in row:
        emitter.EmitVAdd('s32', row_register, row_register, lhs_offset)

    for row in data:
      for (row_register, rhs_offsets_register) in zip(row, self.rhs_offsets):
        emitter.EmitVAdd('s32', row_register, row_register,
                         rhs_offsets_register)

    return data

  def Type(self):
    return 32


class _StaticQuantizationFloatTransformation(object):
  """."""

  def Prepare(self, emitter, registers, kernel_m, kernel_n, lhs, rhs):
    emitter.EmitNewline()
    emitter.EmitComment('StaticQuantizationFloat::Prepare')

    lhs_offset = _ReadParams(emitter, registers, lhs, kernel_m, 4)
    self.rhs_offsets = _ReadParams(emitter, registers, rhs, kernel_n, 4)
    self.scale = _DuplicateGeneralRegister(
        emitter, registers,
        registers.MapParameter('scale', 'params.kernel.scale'), 4)
    self.lhs_offsets = _Duplicate(emitter, registers, kernel_m, lhs_offset)

  def Transform(self, emitter, unused_registers, data, unused_kernel_m,
                unused_kernel_n):
    """Quantize data and output as float."""
    emitter.EmitNewline()
    emitter.EmitComment('StaticQuantizationFloat::Transform')

    for (row, lhs_offset) in zip(data, self.lhs_offsets):
      for row_register in row:
        emitter.EmitVAdd('s32', row_register, row_register, lhs_offset)

    for row in data:
      for (row_register, rhs_offsets_register) in zip(row, self.rhs_offsets):
        emitter.EmitVAdd('s32', row_register, row_register,
                         rhs_offsets_register)

    for row in data:
      for row_register in row:
        emitter.EmitVCvt('f32', 's32', row_register, row_register)

    for row in data:
      for row_register in row:
        emitter.EmitVMul('f32', row_register, row_register, self.scale)

    return data

  def Type(self):
    return 32


class _RowMajorOutput(object):
  """Output data in row major layout."""

  def Prepare(self, emitter, registers, kernel_m, unused_kernel_n,
              unused_data_type):
    """Prepare strided load addresses."""
    emitter.EmitNewline()
    emitter.EmitComment('RowMajorOutput::Prepare')

    stride = registers.MapParameter('stride', 'params.output_stream.stride')

    self.outputs = []
    self.outputs.append(registers.MapOutputParameter('result'))

    for unused_i in range(kernel_m - 1):
      register = registers.GeneralRegister()
      emitter.EmitAdd(register, self.outputs[-1], stride)
      self.outputs.append(register)

  def Output(self, emitter, unused_registers, data, data_type, unused_kernel_m,
             kernel_n):
    emitter.EmitNewline()
    emitter.EmitComment('RowMajorOutput::Output')

    for (datum, output) in zip(data, self.outputs):
      emitter.EmitVStoreAE(data_type, kernel_n, datum, output, None)


def _GenerateAndClearAggregators(emitter, registers, count):
  """Prepare aggregators and emit aggregator clear code."""
  emitter.EmitNewline()
  emitter.EmitComment('Clear aggregators.')
  aggregators = [registers.QuadRegister() for unused_i in range(count)]
  for i in range(count):
    if i < 3:
      emitter.EmitVMov('i32', aggregators[i], emitter.ImmediateConstant(0))
    else:
      emitter.EmitVMov('i32', aggregators[i], aggregators[i - 3])
  return aggregators


def _Generate3x3LoadMultiplyAggregate(emitter, registers, aggregators, lhs, rhs,
                                      count):
  """Emit inner loop for 3 rows x 3 cols multiplication."""
  emitter.EmitNewline()
  emitter.EmitComment('3x3 lanes loop.')
  emitter.EmitNumericalLabel(1)
  emitter.EmitNewline()

  lhs_load = [registers.DoubleRegister() for unused_i in range(3)]
  rhs_load = [registers.DoubleRegister() for unused_i in range(3)]
  temp = [registers.QuadRegister() for unused_i in range(4)]

  emitter.EmitVLoadA(1, 8, rhs_load, emitter.DereferenceIncrement(rhs, 64))
  emitter.EmitVLoad(1, 8, lhs_load[0], emitter.DereferenceIncrement(lhs, 64))

  emitter.EmitVMull('u8', temp[0], lhs_load[0], rhs_load[0])
  emitter.EmitVLoad(1, 8, lhs_load[1], emitter.DereferenceIncrement(lhs, 64))

  emitter.EmitVMull('u8', temp[1], lhs_load[0], rhs_load[1])
  emitter.EmitVLoad(1, 8, lhs_load[2], emitter.DereferenceIncrement(lhs, 64))

  emitter.EmitVMull('u8', temp[2], lhs_load[0], rhs_load[2])
  emitter.EmitPldOffset(lhs, emitter.ImmediateConstant(64))

  emitter.EmitVMull('u8', temp[3], lhs_load[1], rhs_load[0])
  emitter.EmitPldOffset(rhs, emitter.ImmediateConstant(64))

  emitter.EmitVPadal('u16', aggregators[0], temp[0])
  emitter.EmitVPadal('u16', aggregators[1], temp[1])
  emitter.EmitVPadal('u16', aggregators[2], temp[2])
  emitter.EmitVPadal('u16', aggregators[3], temp[3])

  emitter.EmitVMull('u8', temp[0], lhs_load[1], rhs_load[1])
  emitter.EmitVMull('u8', temp[1], lhs_load[1], rhs_load[2])

  registers.FreeRegisters([lhs_load[0], lhs_load[1]])
  temp.append(registers.QuadRegister())

  emitter.EmitVMull('u8', temp[2], lhs_load[2], rhs_load[0])
  emitter.EmitVMull('u8', temp[3], lhs_load[2], rhs_load[1])

  emitter.EmitNewline()
  emitter.EmitComment('Subtract counter.')
  emitter.EmitSubs(count, count, emitter.ImmediateConstant(8))
  emitter.EmitNewline()

  emitter.EmitVMull('u8', temp[4], lhs_load[2], rhs_load[2])

  emitter.EmitVPadal('u16', aggregators[4], temp[0])
  emitter.EmitVPadal('u16', aggregators[5], temp[1])
  emitter.EmitVPadal('u16', aggregators[6], temp[2])
  emitter.EmitVPadal('u16', aggregators[7], temp[3])
  emitter.EmitVPadal('u16', aggregators[8], temp[4])

  emitter.EmitNewline()
  emitter.EmitComment('Loop break.')
  emitter.EmitBgtBack(1)

  registers.FreeRegisters(temp + [lhs_load[2]] + rhs_load)


def _Generate2x4LoadMultiplyAggregate(emitter, registers, aggregators, lhs, rhs,
                                      count):
  """Emit inner loop for 2 rows x 4 cols multiplication."""
  emitter.EmitNewline()
  emitter.EmitComment('2x4 lanes loop.')
  emitter.EmitNumericalLabel(1)
  emitter.EmitNewline()

  lhs_load = [registers.DoubleRegister() for unused_i in range(2)]
  rhs_load = [registers.DoubleRegister() for unused_i in range(4)]
  temp = [registers.QuadRegister() for unused_i in range(5)]

  emitter.EmitVLoadA(1, 8, rhs_load, emitter.DereferenceIncrement(rhs, 256))
  emitter.EmitVLoad(1, 8, lhs_load[0], emitter.DereferenceIncrement(lhs, 64))

  emitter.EmitVMull('u8', temp[0], lhs_load[0], rhs_load[0])
  emitter.EmitVLoad(1, 8, lhs_load[1], emitter.DereferenceIncrement(lhs, 64))

  emitter.EmitVMull('u8', temp[1], lhs_load[0], rhs_load[1])
  emitter.EmitPldOffset(rhs, emitter.ImmediateConstant(64))

  emitter.EmitVMull('u8', temp[2], lhs_load[0], rhs_load[2])
  emitter.EmitPldOffset(lhs, emitter.ImmediateConstant(64))

  emitter.EmitVMull('u8', temp[3], lhs_load[0], rhs_load[3])
  emitter.EmitVMull('u8', temp[4], lhs_load[1], rhs_load[0])

  emitter.EmitVPadal('u16', aggregators[0], temp[0])
  emitter.EmitVPadal('u16', aggregators[1], temp[1])
  emitter.EmitVPadal('u16', aggregators[2], temp[2])

  emitter.EmitVMull('u8', temp[0], lhs_load[1], rhs_load[1])
  emitter.EmitVMull('u8', temp[1], lhs_load[1], rhs_load[2])
  emitter.EmitVMull('u8', temp[2], lhs_load[1], rhs_load[3])

  emitter.EmitNewline()
  emitter.EmitComment('Subtract counter.')
  emitter.EmitSubs(count, count, emitter.ImmediateConstant(8))

  emitter.EmitNewline()
  emitter.EmitVPadal('u16', aggregators[3], temp[3])
  emitter.EmitVPadal('u16', aggregators[4], temp[4])
  emitter.EmitVPadal('u16', aggregators[5], temp[0])
  emitter.EmitVPadal('u16', aggregators[6], temp[1])
  emitter.EmitVPadal('u16', aggregators[7], temp[2])

  emitter.EmitNewline()
  emitter.EmitComment('Loop break.')
  emitter.EmitBgtBack(1)

  registers.FreeRegisters(temp + lhs_load + rhs_load)


def _Generate1x8LoadMultiplyAggregate(emitter, registers, aggregators, lhs, rhs,
                                      count):
  """Emit inner loop for 1 rows x 8 cols multiplication."""
  emitter.EmitNewline()
  emitter.EmitComment('1x8 lanes loop.')
  emitter.EmitNumericalLabel(1)
  emitter.EmitNewline()

  lhs_load = registers.DoubleRegister()
  rhs_load = [registers.DoubleRegister() for unused_i in range(4)]
  temp = [registers.QuadRegister() for unused_i in range(5)]

  emitter.EmitVLoadAE(4 * 8, 8, rhs_load, rhs, 256)
  emitter.EmitVLoadE(8, 8, lhs_load, lhs, 64)

  emitter.EmitVMull('u8', temp[0], lhs_load, rhs_load[0])
  emitter.EmitVMull('u8', temp[1], lhs_load, rhs_load[1])
  emitter.EmitVMull('u8', temp[2], lhs_load, rhs_load[2])
  emitter.EmitVMull('u8', temp[3], lhs_load, rhs_load[3])

  emitter.EmitVLoadAE(4 * 8, 8, rhs_load, rhs, 256)

  emitter.EmitVPadal('u16', aggregators[0], temp[0])
  emitter.EmitVPadal('u16', aggregators[1], temp[1])
  emitter.EmitVPadal('u16', aggregators[2], temp[2])
  emitter.EmitVPadal('u16', aggregators[3], temp[3])

  emitter.EmitPldOffset(rhs, emitter.ImmediateConstant(256))

  emitter.EmitVMull('u8', temp[4], lhs_load, rhs_load[0])
  emitter.EmitVMull('u8', temp[0], lhs_load, rhs_load[1])
  emitter.EmitVMull('u8', temp[1], lhs_load, rhs_load[2])
  emitter.EmitVMull('u8', temp[2], lhs_load, rhs_load[3])

  emitter.EmitPldOffset(lhs, emitter.ImmediateConstant(32))

  emitter.EmitNewline()
  emitter.EmitComment('Subtract counter.')
  emitter.EmitSubs(count, count, emitter.ImmediateConstant(8))

  emitter.EmitNewline()
  emitter.EmitVPadal('u16', aggregators[4], temp[4])
  emitter.EmitVPadal('u16', aggregators[5], temp[0])
  emitter.EmitVPadal('u16', aggregators[6], temp[1])
  emitter.EmitVPadal('u16', aggregators[7], temp[2])

  emitter.EmitNewline()
  emitter.EmitComment('Loop break.')
  emitter.EmitBgtBack(1)

  registers.FreeRegisters(temp + [lhs_load] + rhs_load)


def _GenerateNxMLoadMultiplyAggregate(emitter, registers, kernel_m, kernel_n,
                                      aggregators, lhs, rhs, count):
  """Emit inner loop for N rows x M cols multiplication."""
  emitter.EmitNewline()
  emitter.EmitComment('General NxM lanes loop.')
  emitter.EmitNumericalLabel(1)
  emitter.EmitNewline()
  emitter.EmitComment('Subtract counter.')
  emitter.EmitSubs(count, count, emitter.ImmediateConstant(8))
  emitter.EmitNewline()

  lhs_load = [registers.DoubleRegister() for unused_i in range(kernel_m)]
  rhs_load = [registers.DoubleRegister() for unused_i in range(kernel_n)]

  emitter.EmitVLoadAE(8 * kernel_m, 8, lhs_load, lhs, 64)
  emitter.EmitVLoadAE(8 * kernel_n, 8, rhs_load, rhs, 64)

  emitter.EmitPldOffset(lhs, emitter.ImmediateConstant(64))
  emitter.EmitPldOffset(rhs, emitter.ImmediateConstant(64))

  results = [
      registers.QuadRegister() for unused_i in range(kernel_m * kernel_n)
  ]

  for row in range(kernel_m):
    for col in range(kernel_n):
      index = row * kernel_n + col
      emitter.EmitVMull('u8', results[index], rhs_load[col], lhs_load[row])

  for i in range(kernel_m * kernel_n):
    emitter.EmitVPadal('u16', aggregators[i], results[i])

  emitter.EmitNewline()
  emitter.EmitComment('Loop break.')
  emitter.EmitBgtBack(1)

  registers.FreeRegisters(lhs_load + rhs_load + results)


def _Generate1xNLoadMultiplyAggregate(emitter, registers, kernel_n, aggregators,
                                      lhs, rhs, count):
  """Emit inner loop for 1 row x M cols multiplication."""
  assert kernel_n in [5, 6, 7, 8]
  emitter.EmitNewline()
  emitter.EmitComment('General 1xM lanes loop.')
  emitter.EmitNumericalLabel(1)
  emitter.EmitNewline()
  emitter.EmitComment('Subtract counter.')
  emitter.EmitSubs(count, count, emitter.ImmediateConstant(8))
  emitter.EmitNewline()

  leftover = kernel_n - 4

  rhs_load = [registers.DoubleRegister() for unused_i in range(4)]
  lhs_load = registers.DoubleRegister()

  emitter.EmitVLoadAE(8 * 4, 8, rhs_load, rhs, 64)
  emitter.EmitVLoadE(8, 8, lhs_load, lhs, 64)

  emitter.EmitPldOffset(lhs, emitter.ImmediateConstant(64))

  results = [registers.QuadRegister() for unused_i in range(4)]

  for i in range(4):
    emitter.EmitVMull('u8', results[i], rhs_load[i], lhs_load)

  emitter.EmitVLoadAE(8 * leftover, 8, rhs_load, rhs, 64)
  emitter.EmitPldOffset(rhs, emitter.ImmediateConstant(128))

  for i in range(4):
    emitter.EmitVPadal('u16', aggregators[i], results[i])

  for i in range(leftover):
    emitter.EmitVMull('u8', results[i], rhs_load[i], lhs_load)

  for i in range(leftover):
    emitter.EmitVPadal('u16', aggregators[i + 4], results[i])

  emitter.EmitNewline()
  emitter.EmitComment('Loop break.')
  emitter.EmitBgtBack(1)

  registers.FreeRegisters([lhs_load] + rhs_load + results)


def _GenerateMultiplyKernel(emitter, registers, kernel_m, kernel_n, lhs, rhs):
  """Main muliply loop. Pick best implementation for given kernel shape."""
  count = registers.MapParameter('count', 'params.kernel.count')

  aggregators = _GenerateAndClearAggregators(emitter, registers,
                                             kernel_m * kernel_n)
  if kernel_m is 3 and kernel_n is 3:
    _Generate3x3LoadMultiplyAggregate(emitter, registers, aggregators, lhs, rhs,
                                      count)
  elif kernel_m is 2 and kernel_n is 4:
    _Generate2x4LoadMultiplyAggregate(emitter, registers, aggregators, lhs, rhs,
                                      count)
  elif kernel_m is 1 and kernel_n is 8:
    _Generate1x8LoadMultiplyAggregate(emitter, registers, aggregators, lhs, rhs,
                                      count)
  elif kernel_m is 1 and kernel_n > 4:
    _Generate1xNLoadMultiplyAggregate(emitter, registers, kernel_n, aggregators,
                                      lhs, rhs, count)
  else:
    _GenerateNxMLoadMultiplyAggregate(emitter, registers, kernel_m, kernel_n,
                                      aggregators, lhs, rhs, count)
  return aggregators


def _ReduceAggregators(emitter, aggregators):
  reduced_count = (len(aggregators) + 3) / 4
  reduced = aggregators[:reduced_count]
  emitter.EmitVSumReduce('u32', len(aggregators), 4, reduced, aggregators)
  return reduced


def _GenerateAggregatorReduce(emitter, aggregators, kernel_m, kernel_n):
  emitter.EmitNewline()
  emitter.EmitComment('Reduce aggregators.')
  row_temps = []
  for i in range(kernel_m):
    row_temps.append(
        _ReduceAggregators(emitter, aggregators[i * kernel_n:(i + 1) *
                                                kernel_n]))
  return row_temps


class QuantizedMulKernel(common.MulKernelGenerator):
  """."""

  def __init__(self, cc_emitter, kernel_name, output_stream_name, asm_emitter,
               fused_transformation, output_strategy):
    common.MulKernelGenerator.__init__(self, cc_emitter, kernel_name,
                                       output_stream_name)
    self.asm_emitter = asm_emitter
    self.fused_transformation = fused_transformation
    self.output_strategy = output_strategy

  def EmitMultiply(self, in_type, out_type, kernel_m, kernel_n, pack_size):
    assert in_type is 'uint8_t'
    assert pack_size is 8
    assert kernel_m * kernel_n <= 9

    registers = self.asm_emitter.CreateRegisters()

    self.asm_emitter.PushIndent(self.emitter.indent)
    self.asm_emitter.EmitAsmBegin()

    lhs = registers.MapOutputParameter('lhs')
    rhs = registers.MapOutputParameter('rhs')
    self.asm_emitter.EmitPld(lhs)
    self.asm_emitter.EmitPld(rhs)

    aggregators = _GenerateMultiplyKernel(self.asm_emitter, registers, kernel_m,
                                          kernel_n, lhs, rhs)

    self.fused_transformation.Prepare(self.asm_emitter, registers, kernel_m,
                                      kernel_n, lhs, rhs)

    self.output_strategy.Prepare(self.asm_emitter, registers, kernel_m,
                                 kernel_n, self.fused_transformation.Type())

    reduced = _GenerateAggregatorReduce(self.asm_emitter, aggregators, kernel_m,
                                        kernel_n)

    transformed = self.fused_transformation.Transform(self.asm_emitter,
                                                      registers, reduced,
                                                      kernel_m, kernel_n)

    self.output_strategy.Output(self.asm_emitter, registers, transformed,
                                self.fused_transformation.Type(), kernel_m,
                                kernel_n)

    self.asm_emitter.EmitAsmEnd(registers)
    self.asm_emitter.PopIndent(len(self.emitter.indent))


class QuantizedMulStaticRowMajor(QuantizedMulKernel):
  """."""

  def __init__(self, cc_emitter, asm_emitter):
    QuantizedMulKernel.__init__(self, cc_emitter, 'QuantizedStaticPreprocessed',
                                'RowMajor', asm_emitter,
                                _StaticQuantizationUInt8Transformation(),
                                _RowMajorOutput())


class QuantizedMulStaticAsInt32RowMajor(QuantizedMulKernel):
  """."""

  def __init__(self, cc_emitter, asm_emitter):
    QuantizedMulKernel.__init__(self, cc_emitter,
                                'QuantizedStaticPreprocessedAsInt32',
                                'RowMajor', asm_emitter,
                                _StaticQuantizationInt32Transformation(),
                                _RowMajorOutput())


class QuantizedMulStaticAsFloatRowMajor(QuantizedMulKernel):
  """."""

  def __init__(self, cc_emitter, asm_emitter):
    QuantizedMulKernel.__init__(self, cc_emitter,
                                'QuantizedStaticPreprocessedAsFloat',
                                'RowMajor', asm_emitter,
                                _StaticQuantizationFloatTransformation(),
                                _RowMajorOutput())


def GenerateKernels(cc_emitter, asm_emitter, shapes):
  """Generate the quantized multiplication kernels for uint8 operands."""
  quantized_mul_static_row_major = QuantizedMulStaticRowMajor(cc_emitter,
                                                              asm_emitter)
  quantized_mul_static_int32_row_major = QuantizedMulStaticAsInt32RowMajor(
      cc_emitter, asm_emitter)

  quantized_mul_static_float_row_major = QuantizedMulStaticAsFloatRowMajor(
      cc_emitter, asm_emitter)

  for shape in shapes:
    quantized_mul_static_row_major.SpecializeMulKernel('uint8_t', 'uint8_t',
                                                       shape[0], shape[1], 8)
  for shape in shapes:
    quantized_mul_static_int32_row_major.SpecializeMulKernel('uint8_t',
                                                             'int32_t',
                                                             shape[0], shape[1],
                                                             8)

  for shape in shapes:
    quantized_mul_static_float_row_major.SpecializeMulKernel('uint8_t', 'float',
                                                             shape[0], shape[1],
                                                             8)



# Copyright 2016 The Gemmlowp Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""."""

import common


def _DuplicateGeneralRegister(size, emitter, registers, value, min_register):
  register = registers.QuadRegister(min_register)
  emitter.EmitVDup(size, register, value)
  return register


def _DuplicateGeneralMemoryRegister(size, emitter, registers, value,
                                    min_register):
  register = registers.QuadRegister(min_register)
  general = registers.GeneralRegister()
  emitter.EmitLdr(general, value)
  emitter.EmitVDup(size, register, general)
  registers.FreeRegister(general)
  return register


class MinMaxTransformation(object):
  """."""

  def Check(self, in_type, out_type, kernel_size, leftovers):
    assert in_type is 'uint8_t'
    assert out_type is 'uint8_t'
    assert kernel_size is 16
    assert leftovers < 16

  def Prepare(self, emitter, registers, unused_kernel_size):
    emitter.EmitNewline()
    emitter.EmitComment('MinMax::Prepare')

    self.min = _DuplicateGeneralRegister(8, emitter, registers,
                                         registers.MapParameter('min',
                                                                'params.min'),
                                         4)
    self.max = _DuplicateGeneralRegister(8, emitter, registers,
                                         registers.MapParameter('max',
                                                                'params.max'),
                                         4)

  def Transform(self, emitter, registers, input_address, elements,
                output_address):
    """Generate the MinMax transform inner loop code."""
    emitter.EmitNewline()
    emitter.EmitComment('MinMax::Transform')
    register_count = (elements + 15) / 16
    load = [registers.QuadRegister() for unused_i in range(register_count)]
    emitter.EmitVLoadAE(8, elements, load, input_address, None)
    emitter.EmitPldOffset(input_address, emitter.ImmediateConstant(16))

    for register in load:
      emitter.EmitVMax('u8', register, register, self.min)

    for register in load:
      emitter.EmitVMin('u8', register, register, self.max)

    emitter.EmitNewline()
    emitter.EmitVStoreAE(8, elements, load, output_address, None)
    emitter.EmitPld(output_address)
    registers.FreeRegisters(load)


class DequantizeTransformation(object):
  """."""

  def Check(self, in_type, out_type, kernel_size, leftovers):
    assert in_type is 'uint8_t'
    assert out_type is 'float'
    assert kernel_size is 16
    assert leftovers < 16

  def Prepare(self, emitter, registers, unused_kernel_size):
    """Duplicate quantization offsets to vector registers."""
    emitter.EmitNewline()
    emitter.EmitComment('Dequantize::Prepare')

    self.range_min = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('range_min', 'params.range_min'), 4)
    self.range_offset = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('range_offset', 'params.range_offset'), 4)
    self.range_scale = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('range_scale', 'params.range_scale'), 4)

  def Transform(self, emitter, registers, input_address, elements,
                output_address):
    """Emit the dequantization inner loop."""
    emitter.EmitNewline()
    emitter.EmitComment('Dequantize::Transform')
    register_count = (elements + 3) / 4
    load = [registers.QuadRegister() for unused_i in range(register_count)]
    emitter.EmitVLoadAE(8, elements, load, input_address, None)
    emitter.EmitPldOffset(input_address, emitter.ImmediateConstant(32))

    if len(load) is 1:
      emitter.EmitVMovl('u8', load[0], load[0])
      emitter.EmitVMovl('s16', load[0], load[0])
    elif len(load) is 2:
      emitter.EmitVMovl('u8', load[0], load[0])
      emitter.EmitVMovl2('s16', load[0], load[1], load[0])
    elif len(load) is 3:
      emitter.EmitVMovl2('u8', load[0], load[1], load[0])
      emitter.EmitVMovl('s16', load[2], load[1])
      emitter.EmitVMovl2('s16', load[0], load[1], load[0])
    elif len(load) is 4:
      emitter.EmitVMovl2('u8', load[0], load[1], load[0])
      emitter.EmitVMovl2('s16', load[2], load[3], load[1])
      emitter.EmitVMovl2('s16', load[0], load[1], load[0])
    else:
      assert False

    for register in load:
      emitter.EmitVCvt('f32', 's32', register, register)

    for register in load:
      emitter.EmitVSub('f32', register, register, self.range_offset)

    for register in load:
      emitter.EmitVMul('f32', register, register, self.range_scale)

    for register in load:
      emitter.EmitVAdd('f32', register, register, self.range_min)

    emitter.EmitNewline()
    emitter.EmitVStoreAE(32, elements, load, output_address, None)
    emitter.EmitPld(output_address)
    registers.FreeRegisters(load)


class QuantizeTransformation(object):
  """."""

  def Check(self, in_type, out_type, kernel_size, leftovers):
    assert in_type is 'float'
    assert out_type is 'uint8_t'
    assert kernel_size is 16
    assert leftovers < 16

  def Prepare(self, emitter, registers, unused_kernel_size):
    """Duplicate quantization offsets to vector registers."""
    emitter.EmitNewline()
    emitter.EmitComment('Quantize::Prepare')

    self.range_min = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('range_min', 'params.range_min'), 4)
    self.range_offset = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('range_offset', 'params.range_offset'), 4)
    self.range_scale = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('range_scale', 'params.range_scale'), 4)

  def Transform(self, emitter, registers, input_address, elements,
                output_address):
    """Emit quantization inner loop code."""
    emitter.EmitNewline()
    emitter.EmitComment('Quantize::Transform')
    register_count = (elements + 3) / 4
    load = [registers.QuadRegister() for unused_i in range(register_count)]
    emitter.EmitVLoadAE(32, elements, load, input_address, None)
    emitter.EmitPldOffset(input_address, emitter.ImmediateConstant(64))

    for register in load:
      emitter.EmitVSub('f32', register, register, self.range_min)

    for register in load:
      emitter.EmitVMul('f32', register, register, self.range_scale)

    for register in load:
      emitter.EmitVAdd('f32', register, register, self.range_offset)

    for register in load:
      emitter.EmitVCvt('s32', 'f32', register, register)

    if len(load) is 1:
      emitter.EmitVQmovn('s32', load[0], load[0])
      emitter.EmitVQmovun('s16', load[0], load[0])
    elif len(load) is 2:
      emitter.EmitVQmovn2('s32', load[0], load[0], load[1])
      emitter.EmitVQmovun('s16', load[0], load[0])
    elif len(load) is 3:
      emitter.EmitVQmovn2('s32', load[0], load[0], load[1])
      emitter.EmitVQmovn('s32', load[2], load[2])
      emitter.EmitVQmovun2('s16', load[0], load[0], load[2])
    elif len(load) is 4:
      emitter.EmitVQmovn2('s32', load[0], load[0], load[1])
      emitter.EmitVQmovn2('s32', load[2], load[2], load[3])
      emitter.EmitVQmovun2('s16', load[0], load[0], load[2])
    else:
      assert False

    emitter.EmitNewline()
    emitter.EmitVStoreAE(8, elements, load, output_address, None)
    emitter.EmitPld(output_address)
    registers.FreeRegisters(load)


class RequantizeTransformation(object):
  """."""

  def Check(self, in_type, out_type, kernel_size, leftovers):
    assert in_type is 'int32_t'
    assert out_type is 'uint8_t'
    assert kernel_size is 16
    assert leftovers < 16

  def Prepare(self, emitter, registers, unused_kernel_size):
    """Duplicate quantization parameters to vector registers."""
    emitter.EmitNewline()
    emitter.EmitComment('Requantize::Prepare')

    self.range_min_delta = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('input_range_min', 'params.input_range_min'), 4)
    self.output_range_min = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('output_range_min', 'params.output_range_min'),
        4)
    self.input_range_offset = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('input_range_offset',
                               'params.input_range_offset'), 4)
    self.input_range_scale = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('input_range_scale', 'params.input_range_scale'),
        4)
    self.one_over_output_range_scale = _DuplicateGeneralRegister(
        32, emitter, registers,
        registers.MapParameter('one_over_output_range_scale',
                               'params.one_over_output_range_scale'), 4)
    emitter.EmitVSub('f32', self.range_min_delta, self.range_min_delta,
                     self.output_range_min)

  def Transform(self, emitter, registers, input_address, elements,
                output_address):
    """Emit requantization inner loop code."""
    emitter.EmitNewline()
    emitter.EmitComment('Requantize::Transform')
    register_count = (elements + 3) / 4
    load = [registers.QuadRegister() for unused_i in range(register_count)]
    emitter.EmitVLoadAE(32, elements, load, input_address, None)
    emitter.EmitPldOffset(input_address, emitter.ImmediateConstant(64))

    for register in load:
      emitter.EmitVCvt('f32', 's32', register, register)

    for register in load:
      emitter.EmitVSub('f32', register, register, self.input_range_offset)

    for register in load:
      emitter.EmitVMul('f32', register, register, self.input_range_scale)

    for register in load:
      emitter.EmitVAdd('f32', register, register, self.range_min_delta)

    for register in load:
      emitter.EmitVMul('f32', register, register,
                       self.one_over_output_range_scale)

    for register in load:
      emitter.EmitVCvt('s32', 'f32', register, register)

    if len(load) is 1:
      emitter.EmitVQmovn('s32', load[0], load[0])
      emitter.EmitVQmovun('s16', load[0], load[0])
    elif len(load) is 2:
      emitter.EmitVQmovn2('s32', load[0], load[0], load[1])
      emitter.EmitVQmovun('s16', load[0], load[0])
    elif len(load) is 3:
      emitter.EmitVQmovn2('s32', load[0], load[0], load[1])
      emitter.EmitVQmovn('s32', load[2], load[2])
      emitter.EmitVQmovun2('s16', load[0], load[0], load[2])
    elif len(load) is 4:
      emitter.EmitVQmovn2('s32', load[0], load[0], load[1])
      emitter.EmitVQmovn2('s32', load[2], load[2], load[3])
      emitter.EmitVQmovun2('s16', load[0], load[0], load[2])
    else:
      assert False

    emitter.EmitNewline()
    emitter.EmitVStoreAE(8, elements, load, output_address, None)
    emitter.EmitPld(output_address)
    registers.FreeRegisters(load)


class BaseTransform(common.Transform1DKernelGenerator):
  """."""

  def __init__(self, cc_emitter, kernel_name, asm_emitter, transformation):
    common.Transform1DKernelGenerator.__init__(self, cc_emitter, kernel_name)
    self.asm_emitter = asm_emitter
    self.transformation = transformation

  def EmitTransform(self, in_type, out_type, kernel_size, leftovers):
    """."""
    self.transformation.Check(in_type, out_type, kernel_size, leftovers)

    registers = self.asm_emitter.CreateRegisters()

    self.emitter.EmitDeclare('int', 'params_count_copy', 'params.count')

    self.asm_emitter.PushIndent(self.emitter.indent)
    self.asm_emitter.EmitAsmBegin()

    count = registers.MapOutputParameter('count', 'params_count_copy')
    input_address = registers.MapOutputParameter('input')
    output_address = registers.MapOutputParameter('output')

    self.transformation.Prepare(self.asm_emitter, registers, kernel_size)

    if leftovers:
      self.asm_emitter.EmitNewline()
      self.asm_emitter.EmitComment('Reduce count by leftovers.')
      self.asm_emitter.EmitSubs(count, count,
                                self.asm_emitter.ImmediateConstant(leftovers))
      self.asm_emitter.EmitBeqFront(2)

    self.asm_emitter.EmitNewline()
    self.asm_emitter.EmitNumericalLabel(1)
    self.asm_emitter.EmitSubs(count, count,
                              self.asm_emitter.ImmediateConstant(kernel_size))

    self.transformation.Transform(self.asm_emitter, registers, input_address,
                                  kernel_size, output_address)

    self.asm_emitter.EmitNewline()
    self.asm_emitter.EmitBneBack(1)

    if leftovers:
      self.asm_emitter.EmitNumericalLabel(2)
      self.asm_emitter.EmitNewline()
      self.asm_emitter.EmitComment('Handle leftovers.')
      self.transformation.Transform(self.asm_emitter, registers, input_address,
                                    leftovers, output_address)

    self.asm_emitter.EmitAsmEnd(registers)
    self.asm_emitter.PopIndent(len(self.emitter.indent))


class Requantize(BaseTransform):
  """."""

  def __init__(self, cc_emitter, asm_emitter):
    BaseTransform.__init__(self, cc_emitter, 'Requantize', asm_emitter,
                           RequantizeTransformation())


class Quantize(BaseTransform):
  """."""

  def __init__(self, cc_emitter, asm_emitter):
    BaseTransform.__init__(self, cc_emitter, 'Quantize', asm_emitter,
                           QuantizeTransformation())


class Dequantize(BaseTransform):
  """."""

  def __init__(self, cc_emitter, asm_emitter):
    BaseTransform.__init__(self, cc_emitter, 'Dequantize', asm_emitter,
                           DequantizeTransformation())


class MinMax(BaseTransform):
  """."""

  def __init__(self, numerical_type, cc_emitter, asm_emitter):
    BaseTransform.__init__(self, cc_emitter, 'MinMax<%s>' % numerical_type,
                           asm_emitter, MinMaxTransformation())


class BiasAdd(common.Transform1DKernelGenerator):
  """."""

  def __init__(self, bias_type, cc_emitter, asm_emitter):
    common.Transform1DKernelGenerator.__init__(self, cc_emitter,
                                               'BiasAdd<%s>' % bias_type)
    self.asm_emitter = asm_emitter

  def EmitTransform(self, in_type, out_type, kernel_size, leftovers):
    """."""
    assert in_type is 'uint8_t'
    assert out_type is 'int32_t'
    assert kernel_size is 16
    assert leftovers < 16

    registers = self.asm_emitter.CreateRegisters()

    self.emitter.EmitDeclare('int', 'params_rows_copy', 'params.rows')

    self.asm_emitter.PushIndent(self.emitter.indent)
    self.asm_emitter.EmitAsmBegin()

    self._Prepare(self.asm_emitter, registers)

    rows = registers.MapParameter('rows', 'params_rows_copy')

    self.asm_emitter.EmitNumericalLabel(1)

    self._ProcessRow(self.asm_emitter, registers, kernel_size, leftovers)

    self.asm_emitter.EmitSubs(rows, rows, self.asm_emitter.ImmediateConstant(1))
    self.asm_emitter.EmitBneBack(1)

    self.asm_emitter.EmitAsmEnd(registers)
    self.asm_emitter.PopIndent(len(self.emitter.indent))

  def _Prepare(self, emitter, registers):
    self.input_range_min = _DuplicateGeneralMemoryRegister(
        32, emitter, registers,
        registers.MapMemoryParameter('input_range_min',
                                     'params.input_range_min'), 8)
    self.input_range_scale = _DuplicateGeneralMemoryRegister(
        32, emitter, registers,
        registers.MapMemoryParameter('input_range_scale',
                                     'params.input_range_scale'), 8)
    self.bias_range_min = _DuplicateGeneralMemoryRegister(
        32, emitter, registers,
        registers.MapMemoryParameter('bias_range_min', 'params.bias_range_min'),
        8)
    self.bias_range_scale = _DuplicateGeneralMemoryRegister(
        32, emitter, registers,
        registers.MapMemoryParameter('bias_range_scale',
                                     'params.bias_range_scale'), 8)
    self.output_range_min = _DuplicateGeneralMemoryRegister(
        32, emitter, registers,
        registers.MapMemoryParameter('output_range_min',
                                     'params.output_range_min'), 8)
    self.one_over_output_range_scale = _DuplicateGeneralMemoryRegister(
        32, emitter, registers,
        registers.MapMemoryParameter('one_over_output_range_scale',
                                     'params.one_over_output_range_scale'), 8)
    self.output_range_offset = _DuplicateGeneralMemoryRegister(
        32, emitter, registers,
        registers.MapMemoryParameter('output_range_offset',
                                     'params.output_range_offset'), 8)

  def _ProcessRow(self, emitter, registers, kernel_size, leftovers):
    const_count = registers.MapParameter('count', 'params.count')
    const_bias = registers.MapParameter('bias', 'params.bias')

    count = registers.GeneralRegister()
    bias = registers.GeneralRegister()

    input_address = registers.MapOutputParameter('input')
    output_address = registers.MapOutputParameter('output')

    emitter.EmitMov(count, const_count)
    emitter.EmitMov(bias, const_bias)

    if leftovers:
      emitter.EmitSubs(count, count, emitter.ImmediateConstant(leftovers))
      emitter.EmitBeqFront(3)

    emitter.EmitNumericalLabel(2)
    emitter.EmitSubs(count, count, emitter.ImmediateConstant(kernel_size))

    self._BiasAdd(emitter, registers, kernel_size, input_address, bias,
                  output_address)

    emitter.EmitBneBack(2)

    if leftovers:
      emitter.EmitNumericalLabel(3)
      self._BiasAdd(emitter, registers, leftovers, input_address, bias,
                    output_address)

  def _BiasAdd(self, emitter, registers, elements, input_address, bias,
               output_address):
    emitter.EmitNewline()
    emitter.EmitComment('BiasAdd::Transform')
    register_count = (elements + 3) / 4

    load_input = [
        registers.QuadRegister() for unused_i in range(register_count)
    ]
    load_bias = [registers.QuadRegister() for unused_i in range(register_count)]

    emitter.EmitVLoadAE(8, elements, load_input, input_address, None)
    emitter.EmitVLoadAE(8, elements, load_bias, bias, None)
    emitter.EmitPldOffset(input_address, emitter.ImmediateConstant(32))

    if len(load_input) is 1:
      emitter.EmitVMovl('u8', load_input[0], load_input[0])
      emitter.EmitVMovl('u8', load_bias[0], load_bias[0])
      emitter.EmitVMovl('s16', load_input[0], load_input[0])
      emitter.EmitVMovl('s16', load_bias[0], load_bias[0])
    elif len(load_input) is 2:
      emitter.EmitVMovl('u8', load_input[0], load_input[0])
      emitter.EmitVMovl('u8', load_bias[0], load_bias[0])
      emitter.EmitVMovl2('s16', load_input[0], load_input[1], load_input[0])
      emitter.EmitVMovl2('s16', load_bias[0], load_bias[1], load_bias[0])
    elif len(load_input) is 3:
      emitter.EmitVMovl2('u8', load_input[0], load_input[1], load_input[0])
      emitter.EmitVMovl2('u8', load_bias[0], load_bias[1], load_bias[0])
      emitter.EmitVMovl('s16', load_input[2], load_input[1])
      emitter.EmitVMovl('s16', load_bias[2], load_bias[1])
      emitter.EmitVMovl2('s16', load_input[0], load_input[1], load_input[0])
      emitter.EmitVMovl2('s16', load_bias[0], load_bias[1], load_bias[0])
    elif len(load_input) is 4:
      emitter.EmitVMovl2('u8', load_input[0], load_input[1], load_input[0])
      emitter.EmitVMovl2('u8', load_bias[0], load_bias[1], load_bias[0])
      emitter.EmitVMovl2('s16', load_input[2], load_input[3], load_input[1])
      emitter.EmitVMovl2('s16', load_bias[2], load_bias[3], load_bias[1])
      emitter.EmitVMovl2('s16', load_input[0], load_input[1], load_input[0])
      emitter.EmitVMovl2('s16', load_bias[0], load_bias[1], load_bias[0])
    else:
      assert False

    for register in load_input + load_bias:
      emitter.EmitVCvt('f32', 's32', register, register)

    for register in load_input:
      emitter.EmitVMul('f32', register, register, self.input_range_scale)

    for register in load_bias:
      emitter.EmitVMul('f32', register, register, self.bias_range_scale)

    for register in load_input:
      emitter.EmitVAdd('f32', register, register, self.input_range_min)

    for register in load_bias:
      emitter.EmitVAdd('f32', register, register, self.bias_range_min)

    for (register_1, register_2) in zip(load_input, load_bias):
      emitter.EmitVAdd('f32', register_1, register_1, register_2)

    for register in load_input:
      emitter.EmitVSub('f32', register, register, self.output_range_min)

    for register in load_input:
      emitter.EmitVMul('f32', register, register,
                       self.one_over_output_range_scale)

    for register in load_input:
      emitter.EmitVAdd('f32', register, register, self.output_range_offset)

    for register in load_input:
      emitter.EmitVCvt('s32', 'f32', register, register)

    emitter.EmitNewline()
    emitter.EmitVStoreAE(32, elements, load_input, output_address, None)
    emitter.EmitPld(output_address)
    registers.FreeRegisters(load_input + load_bias)


def GenerateKernels(cc_emitter, asm_emitter, shapes):
  """Generate the quantization/dequantization/requantization kernels."""
  requantize = Requantize(cc_emitter, asm_emitter)
  quantize = Quantize(cc_emitter, asm_emitter)
  dequantize = Dequantize(cc_emitter, asm_emitter)
  minmax = MinMax('uint8_t', cc_emitter, asm_emitter)
  biasadd = BiasAdd('uint8_t', cc_emitter, asm_emitter)

  for shape in shapes:
    requantize.SpecializeTransform1DKernel('int32_t', 'uint8_t', shape[0],
                                           shape[1])

  for shape in shapes:
    quantize.SpecializeTransform1DKernel('float', 'uint8_t', shape[0], shape[1])

  for shape in shapes:
    dequantize.SpecializeTransform1DKernel('uint8_t', 'float', shape[0],
                                           shape[1])

  for shape in shapes:
    minmax.SpecializeTransform1DKernel('uint8_t', 'uint8_t', shape[0], shape[1])

  for shape in shapes:
    biasadd.SpecializeTransform1DKernel('uint8_t', 'int32_t', shape[0],
                                        shape[1])



# Copyright 2018 The gemmlowp Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Encodes ARM asm code for certain instructions into the corresponding machine code encoding, as a .word directive in the asm code, preserving the original code in a comment.

Reads from stdin, writes to stdout.

Example diff:
-        "udot v16.4s, v4.16b, v0.16b\n"
+        ".word 0x6e809490  // udot v16.4s, v4.16b, v0.16b\n"

The intended use case is to make asm code easier to compile on toolchains that
do not support certain new instructions.
"""

import sys
import re
import argparse


def encode_udot_sdot_vector(line):
  m = re.search(
      r'\b([us])dot[ ]+v([0-9]+)[ ]*\.[ ]*4s[ ]*\,[ ]*v([0-9]+)[ ]*\.[ ]*16b[ ]*\,[ ]*v([0-9]+)[ ]*\.[ ]*16b',
      line)
  if not m:
    return 0, line

  match = m.group(0)
  unsigned = 1 if m.group(1) == 'u' else 0
  accum = int(m.group(2))
  lhs = int(m.group(3))
  rhs = int(m.group(4))
  assert accum >= 0 and accum <= 31
  assert lhs >= 0 and lhs <= 31
  assert rhs >= 0 and rhs <= 31
  mcode = 0x4e809400 | (accum << 0) | (lhs << 5) | (rhs << 16) | (
      unsigned << 29)
  return mcode, match


def encode_udot_sdot_element(line):
  m = re.search(
      r'\b([us])dot[ ]+v([0-9]+)[ ]*\.[ ]*4s[ ]*\,[ ]*v([0-9]+)[ ]*\.[ ]*16b[ ]*\,[ ]*v([0-9]+)[ ]*\.[ ]*4b[ ]*\[([0-9])\]',
      line)
  if not m:
    return 0, line

  match = m.group(0)
  unsigned = 1 if m.group(1) == 'u' else 0
  accum = int(m.group(2))
  lhs = int(m.group(3))
  rhs = int(m.group(4))
  lanegroup = int(m.group(5))
  assert accum >= 0 and accum <= 31
  assert lhs >= 0 and lhs <= 31
  assert rhs >= 0 and rhs <= 31
  assert lanegroup >= 0 and lanegroup <= 3
  l = 1 if lanegroup & 1 else 0
  h = 1 if lanegroup & 2 else 0
  mcode = 0x4f80e000 | (accum << 0) | (lhs << 5) | (rhs << 16) | (l << 21) | (
      h << 11) | (
          unsigned << 29)
  return mcode, match


def encode(line):
  for encode_func in [encode_udot_sdot_vector, encode_udot_sdot_element]:
    mcode, match = encode_func(line)
    if mcode:
      return mcode, match
  return 0, line


def read_existing_encoding(line):
  m = re.search(r'\.word\ (0x[0-9a-f]+)', line)
  if m:
    return int(m.group(1), 16)
  return 0


parser = argparse.ArgumentParser(description='Encode some A64 instructions.')
parser.add_argument(
    '-f',
    '--fix',
    help='fix existing wrong encodings in-place and continue',
    action='store_true')
args = parser.parse_args()

lineno = 0
found_existing_encodings = False
found_error = False
found_fixes = False
for line in sys.stdin:
  lineno = lineno + 1
  mcode, match = encode(line)
  if mcode:
    existing_encoding = read_existing_encoding(line)
    if existing_encoding:
      found_existing_encodings = True
      if mcode != existing_encoding:
        if args.fix:
          line = line.replace('.word 0x%x  // %s' % (existing_encoding, match),
                              '.word 0x%x  // %s' % (mcode, match))
          found_fixes = True
        else:
          sys.stderr.write(
              "Error at line %d: existing encoding 0x%x differs from encoding 0x%x for instruction '%s':\n\n%s\n\n"
              % (lineno, existing_encoding, mcode, match, line))
          found_error = True
    else:
      line = line.replace(match, '.word 0x%x  // %s' % (mcode, match))
  sys.stdout.write(line)
if found_error:
  sys.exit(1)
if found_existing_encodings:
  if found_fixes:
    sys.stderr.write(
        'Note: some instructions that this program is able to encode, were already encoded and their existing encodings didn\'t match the specified asm instructions. Since --fix was passed, these were fixed in-place.\n'
    )
  else:
    sys.stderr.write(
        'Note: some instructions that this program is able to encode, were already encoded. These encodings have been checked.\n'
    )



# Copyright 2016 The Gemmlowp Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CC code emitter.

Used by generators to programatically prepare C++ code. Contains some simple
tools that allow generating nicely indented code and do basic correctness
checking.
"""


class Error(Exception):
  """Module level error."""


class NamespaceError(Error):
  """Invalid namespace operation."""


class HeaderError(Error):
  """Invalid cc header structure."""


class ClassError(Error):
  """Invalid class syntax."""


class CCEmitter(object):
  """Emits c++ code."""

  def __init__(self, debug=False):
    self.indent = ''
    self.debug = debug
    self.namespaces = []
    self.classes = []
    self.header_name = None

  def PushIndent(self):
    self.indent += '  '

  def PopIndent(self):
    self.indent = self.indent[:-2]

  def EmitIndented(self, what):
    print(self.indent + what)

  def EmitNewline(self):
    print('')

  def EmitPreprocessor1(self, op, param):
    print('#%s %s' % (op, param))

  def EmitPreprocessor(self, op):
    print('#%s' % op)

  def EmitInclude(self, include):
    self.EmitPreprocessor1('include', include)

  def EmitAssign(self, variable, value):
    self.EmitBinaryOp(variable, '=', value)

  def EmitAssignIncrement(self, variable, value):
    self.EmitBinaryOp(variable, '+=', value)

  def EmitBinaryOp(self, operand_1, op, operand_2):
    self.EmitCode('%s %s %s' % (operand_1, op, operand_2))

  def EmitCall(self, function, params=None):
    if not params:
      params = []
    self.EmitCode('%s(%s)' % (function, ', '.join(map(str, params))))

  def EmitCode(self, code):
    self.EmitIndented('%s;' % code)

  def EmitCodeNoSemicolon(self, code):
    self.EmitIndented('%s' % code)

  def EmitDeclare(self, decl_type, name, value):
    self.EmitAssign('%s %s' % (decl_type, name), value)

  def EmitAssert(self, assert_expression):
    if self.debug:
      self.EmitCall1('assert', assert_expression)

  def EmitHeaderBegin(self, header_name, includes=None):
    if includes is None:
      includes = []
    if self.header_name:
      raise HeaderError('Header already defined.')
    self.EmitPreprocessor1('ifndef', (header_name + '_H_').upper())
    self.EmitPreprocessor1('define', (header_name + '_H_').upper())
    self.EmitNewline()
    if includes:
      for include in includes:
        self.EmitInclude(include)
      self.EmitNewline()
    self.header_name = header_name

  def EmitHeaderEnd(self):
    if not self.header_name:
      raise HeaderError('Header undefined.')
    self.EmitPreprocessor1('endif',
                           ' // %s' % (self.header_name + '_H_').upper())
    self.header_name = None

  def EmitMemberFunctionBegin(self, class_name, class_template_params,
                              class_specializations, function_name,
                              function_params, return_type):
    """Emit member function of a template/specialized class."""
    if class_template_params or class_specializations:
      self.EmitIndented('template<%s>' % ', '.join(class_template_params))

    if class_specializations:
      class_name += '<%s>' % ', '.join(map(str, class_specializations))

    self.EmitIndented('%s %s::%s(%s) {' % (
        return_type, class_name, function_name,
        ', '.join(['%s %s' % (t, n) for (t, n) in function_params])))
    self.PushIndent()

  def EmitFunctionBegin(self, function_name, params, return_type):
    self.EmitIndented('%s %s(%s) {' %
                      (return_type, function_name,
                       ', '.join(['%s %s' % (t, n) for (t, n) in params])))
    self.PushIndent()

  def EmitFunctionEnd(self):
    self.PopIndent()
    self.EmitIndented('}')
    self.EmitNewline()

  def EmitClassBegin(self, class_name, template_params, specializations,
                     base_classes):
    """Emit class block header."""
    self.classes.append(class_name)
    if template_params or specializations:
      self.EmitIndented('template<%s>' % ', '.join(template_params))

    class_name_extended = class_name
    if specializations:
      class_name_extended += '<%s>' % ', '.join(map(str, specializations))
    if base_classes:
      class_name_extended += ' : ' + ', '.join(base_classes)
    self.EmitIndented('class %s {' % class_name_extended)
    self.PushIndent()

  def EmitClassEnd(self):
    if not self.classes:
      raise ClassError('No class on stack.')
    self.classes.pop()
    self.PopIndent()
    self.EmitIndented('};')
    self.EmitNewline()

  def EmitAccessModifier(self, modifier):
    if not self.classes:
      raise ClassError('No class on stack.')
    self.PopIndent()
    self.EmitIndented(' %s:' % modifier)
    self.PushIndent()

  def EmitNamespaceBegin(self, namespace):
    self.EmitCodeNoSemicolon('namespace %s {' % namespace)
    self.namespaces.append(namespace)

  def EmitNamespaceEnd(self):
    if not self.namespaces:
      raise NamespaceError('No namespace on stack.')
    self.EmitCodeNoSemicolon('}  // namespace %s' % self.namespaces.pop())

  def EmitComment(self, comment):
    self.EmitIndented('// ' + comment)

  def EmitOpenBracket(self, pre_bracket=None):
    if pre_bracket:
      self.EmitIndented('%s {' % pre_bracket)
    else:
      self.EmitIndented('{')
    self.PushIndent()

  def EmitCloseBracket(self):
    self.PopIndent()
    self.EmitIndented('}')

  def EmitSwitch(self, switch):
    self.EmitOpenBracket('switch (%s)' % switch)

  def EmitSwitchEnd(self):
    self.EmitCloseBracket()

  def EmitCase(self, value):
    self.EmitCodeNoSemicolon('case %s:' % value)

  def EmitBreak(self):
    self.EmitCode('break')

  def EmitIf(self, condition):
    self.EmitOpenBracket('if (%s)' % condition)

  def EmitElse(self):
    self.PopIndent()
    self.EmitCodeNoSemicolon('} else {')
    self.PushIndent()

  def EmitEndif(self):
    self.EmitCloseBracket()

  def Scope(self, scope, value):
    return '%s::%s' % (scope, value)



# Copyright 2016 The Gemmlowp Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""64bit ARM/NEON assembly emitter.

Used by code generators to produce ARM assembly with NEON simd code.
Provides tools for easier register management: named register variable
allocation/deallocation, and offers a more procedural/structured approach
to generating assembly.

"""

_WIDE_TYPES = {
    8: 16,
    16: 32,
    32: 64,
    '8': '16',
    '16': '32',
    '32': '64',
    'i8': 'i16',
    'i16': 'i32',
    'i32': 'i64',
    'u8': 'u16',
    'u16': 'u32',
    'u32': 'u64',
    's8': 's16',
    's16': 's32',
    's32': 's64'
}

_NARROW_TYPES = {
    64: 32,
    32: 16,
    16: 8,
    '64': '32',
    '32': '16',
    '16': '8',
    'i64': 'i32',
    'i32': 'i16',
    'i16': 'i8',
    'u64': 'u32',
    'u32': 'u16',
    'u16': 'u8',
    's64': 's32',
    's32': 's16',
    's16': 's8'
}

_TYPE_BITS = {
    8: 8,
    16: 16,
    32: 32,
    64: 64,
    '8': 8,
    '16': 16,
    '32': 32,
    '64': 64,
    'i8': 8,
    'i16': 16,
    'i32': 32,
    'i64': 64,
    'u8': 8,
    'u16': 16,
    'u32': 32,
    'u64': 64,
    's8': 8,
    's16': 16,
    's32': 32,
    's64': 64,
    'f32': 32,
    'f64': 64,
    'b': 8,
    'h': 16,
    's': 32,
    'd': 64
}


class Error(Exception):
  """Module level error."""


class RegisterAllocationError(Error):
  """Cannot alocate registers."""


class LaneError(Error):
  """Wrong lane number."""


class RegisterSubtypeError(Error):
  """The register needs to be lane-typed."""


class ArgumentError(Error):
  """Wrong argument."""


def _AppendType(type_name, register):
  """Calculates sizes and attaches the type information to the register."""
  if register.register_type is not 'v':
    raise ArgumentError('Only vector registers can have type appended.')

  if type_name in set([8, '8', 'i8', 's8', 'u8']):
    subtype = 'b'
    subtype_bits = 8
  elif type_name in set([16, '16', 'i16', 's16', 'u16']):
    subtype = 'h'
    subtype_bits = 16
  elif type_name in set([32, '32', 'i32', 's32', 'u32', 'f32']):
    subtype = 's'
    subtype_bits = 32
  elif type_name in set([64, '64', 'i64', 's64', 'u64', 'f64']):
    subtype = 'd'
    subtype_bits = 64
  else:
    raise ArgumentError('Unknown type: %s' % type_name)

  new_register = register.Copy()
  new_register.register_subtype = subtype
  new_register.register_subtype_count = register.register_bits / subtype_bits
  return new_register


def _UnsignedType(type_name):
  return type_name in set(['u8', 'u16', 'u32', 'u64'])


def _FloatType(type_name):
  return type_name in set(['f32', 'f64'])


def _WideType(type_name):
  if type_name in _WIDE_TYPES.keys():
    return _WIDE_TYPES[type_name]
  else:
    raise ArgumentError('No wide type for: %s' % type_name)


def _NarrowType(type_name):
  if type_name in _NARROW_TYPES.keys():
    return _NARROW_TYPES[type_name]
  else:
    raise ArgumentError('No narrow type for: %s' % type_name)


def _LoadStoreSize(register):
  if register.lane is None:
    return register.register_bits
  else:
    return register.lane_bits


def _MakeCompatibleDown(reg_1, reg_2, reg_3):
  bits = min([reg_1.register_bits, reg_2.register_bits, reg_3.register_bits])
  return (_Cast(bits, reg_1), _Cast(bits, reg_2), _Cast(bits, reg_3))


def _MakeCompatibleUp(reg_1, reg_2, reg_3):
  bits = max([reg_1.register_bits, reg_2.register_bits, reg_3.register_bits])
  return (_Cast(bits, reg_1), _Cast(bits, reg_2), _Cast(bits, reg_3))


def _Cast(bits, reg):
  if reg.register_bits is bits:
    return reg
  else:
    new_reg = reg.Copy()
    new_reg.register_bits = bits
    return new_reg


def _TypeBits(type_name):
  if type_name in _TYPE_BITS.keys():
    return _TYPE_BITS[type_name]
  else:
    raise ArgumentError('Unknown type: %s' % type_name)


def _RegisterList(list_type, registers):
  lanes = list(set([register.lane for register in registers]))
  if len(lanes) > 1:
    raise ArgumentError('Cannot mix lanes on a register list.')
  typed_registers = [_AppendType(list_type, register) for register in registers]

  if lanes[0] is None:
    return '{%s}' % ', '.join(map(str, typed_registers))
  elif lanes[0] is -1:
    raise ArgumentError('Cannot construct a list with all lane indexing.')
  else:
    typed_registers_nolane = [register.Copy() for register in typed_registers]
    for register in typed_registers_nolane:
      register.lane = None
      register.register_subtype_count = None
    return '{%s}[%d]' % (', '.join(map(str, typed_registers_nolane)), lanes[0])


class _GeneralRegister(object):
  """Arm v8 general register: (x|w)n."""

  def __init__(self,
               register_bits,
               number,
               dereference=False,
               dereference_increment=False):
    self.register_type = 'r'
    self.register_bits = register_bits
    self.number = number
    self.dereference = dereference
    self.dereference_increment = dereference_increment

  def Copy(self):
    return _GeneralRegister(self.register_bits, self.number, self.dereference,
                            self.dereference_increment)

  def __repr__(self):
    if self.register_bits is 64:
      text = 'x%d' % self.number
    elif self.register_bits <= 32:
      text = 'w%d' % self.number
    else:
      raise RegisterSubtypeError('Wrong bits (%d) for general register: %d' %
                                 (self.register_bits, self.number))
    if self.dereference:
      return '[%s]' % text
    else:
      return text


class _MappedParameter(object):
  """Object representing a C variable mapped to a register."""

  def __init__(self,
               name,
               register_bits=64,
               dereference=False,
               dereference_increment=False):
    self.name = name
    self.register_bits = register_bits
    self.dereference = dereference
    self.dereference_increment = dereference_increment

  def Copy(self):
    return _MappedParameter(self.name, self.register_bits, self.dereference,
                            self.dereference_increment)

  def __repr__(self):
    if self.register_bits is None:
      text = '%%[%s]' % self.name
    elif self.register_bits is 64:
      text = '%%x[%s]' % self.name
    elif self.register_bits <= 32:
      text = '%%w[%s]' % self.name
    else:
      raise RegisterSubtypeError('Wrong bits (%d) for mapped parameter: %s' %
                                 (self.register_bits, self.name))
    if self.dereference:
      return '[%s]' % text
    else:
      return text


class _VectorRegister(object):
  """Arm v8 vector register Vn.TT."""

  def __init__(self,
               register_bits,
               number,
               register_subtype=None,
               register_subtype_count=None,
               lane=None,
               lane_bits=None):
    self.register_type = 'v'
    self.register_bits = register_bits
    self.number = number
    self.register_subtype = register_subtype
    self.register_subtype_count = register_subtype_count
    self.lane = lane
    self.lane_bits = lane_bits

  def Copy(self):
    return _VectorRegister(self.register_bits, self.number,
                           self.register_subtype, self.register_subtype_count,
                           self.lane, self.lane_bits)

  def __repr__(self):
    if self.register_subtype is None:
      raise RegisterSubtypeError('Register: %s%d has no lane types defined.' %
                                 (self.register_type, self.number))
    if (self.register_subtype_count is None or (self.lane is not None and
                                                self.lane is not -1)):
      typed_name = '%s%d.%s' % (self.register_type, self.number,
                                self.register_subtype)
    else:
      typed_name = '%s%d.%d%s' % (self.register_type, self.number,
                                  self.register_subtype_count,
                                  self.register_subtype)

    if self.lane is None or self.lane is -1:
      return typed_name
    elif self.lane >= 0 and self.lane < self.register_subtype_count:
      return '%s[%d]' % (typed_name, self.lane)
    else:
      raise LaneError('Wrong lane: %d for: %s' % (self.lane, typed_name))


class _ImmediateConstant(object):

  def __init__(self, value):
    self.register_type = 'i'
    self.value = value

  def Copy(self):
    return _ImmediateConstant(self.value)

  def __repr__(self):
    return '#%d' % self.value


class _NeonRegisters64Bit(object):
  """Utility that keeps track of used 32bit ARM/NEON registers."""

  def __init__(self):
    self.vector = set()
    self.vector_ever = set()
    self.general = set()
    self.general_ever = set()
    self.parameters = dict()
    self.output_parameters = dict()

  def MapParameter(self, parameter, parameter_value=None):
    if not parameter_value:
      parameter_value = parameter
    self.parameters[parameter] = (parameter_value, 'r')
    return _MappedParameter(parameter)

  def MapMemoryParameter(self, parameter, parameter_value=None):
    if not parameter_value:
      parameter_value = parameter
    self.parameters[parameter] = (parameter_value, 'm')
    return _MappedParameter(parameter)

  def MapOutputParameter(self, parameter, parameter_value=None):
    if not parameter_value:
      parameter_value = parameter
    self.output_parameters[parameter] = (parameter_value, '+r')
    return _MappedParameter(parameter)

  def _VectorRegisterNum(self, min_val=0):
    for i in range(min_val, 32):
      if i not in self.vector:
        self.vector.add(i)
        self.vector_ever.add(i)
        return i
    raise RegisterAllocationError('Not enough vector registers.')

  def DoubleRegister(self, min_val=0):
    return _VectorRegister(64, self._VectorRegisterNum(min_val))

  def QuadRegister(self, min_val=0):
    return _VectorRegister(128, self._VectorRegisterNum(min_val))

  def GeneralRegister(self):
    for i in range(0, 30):
      if i not in self.general:
        self.general.add(i)
        self.general_ever.add(i)
        return _GeneralRegister(64, i)
    raise RegisterAllocationError('Not enough general registers.')

  def MappedParameters(self):
    return [x for x in self.parameters.items()]

  def MappedOutputParameters(self):
    return [x for x in self.output_parameters.items()]

  def Clobbers(self):
    return (
        ['x%d' % i
         for i in self.general_ever] + ['v%d' % i for i in self.vector_ever])

  def FreeRegister(self, register):
    if isinstance(register, _MappedParameter):
      return

    if register.register_type == 'v':
      assert register.number in self.vector
      self.vector.remove(register.number)
    elif register.register_type == 'r':
      assert register.number in self.general
      self.general.remove(register.number)
    else:
      raise RegisterAllocationError('Register not allocated: %s%d' %
                                    (register.register_type, register.number))

  def FreeRegisters(self, registers):
    for register in registers:
      self.FreeRegister(register)


class NeonEmitter64(object):
  """Emits ARM/NEON 64bit assembly opcodes."""

  def __init__(self, debug=False):
    self.ops = {}
    self.indent = ''
    self.debug = debug

  def PushIndent(self, delta_indent='  '):
    self.indent += delta_indent

  def PopIndent(self, delta=2):
    self.indent = self.indent[:-delta]

  def EmitIndented(self, what):
    print(self.indent + what)

  def PushOp(self, op):
    if op in self.ops.keys():
      self.ops[op] += 1
    else:
      self.ops[op] = 1

  def ClearCounters(self):
    self.ops.clear()

  def EmitNewline(self):
    print('')

  def EmitPreprocessor1(self, op, param):
    print('#%s %s' % (op, param))

  def EmitPreprocessor(self, op):
    print('#%s' % op)

  def EmitInclude(self, include):
    self.EmitPreprocessor1('include', include)

  def EmitCall1(self, function, param):
    self.EmitIndented('%s(%s);' % (function, param))

  def EmitAssert(self, assert_expression):
    if self.debug:
      self.EmitCall1('assert', assert_expression)

  def EmitHeaderBegin(self, header_name, includes):
    self.EmitPreprocessor1('ifndef', (header_name + '_H_').upper())
    self.EmitPreprocessor1('define', (header_name + '_H_').upper())
    self.EmitNewline()
    if includes:
      for include in includes:
        self.EmitInclude(include)
      self.EmitNewline()

  def EmitHeaderEnd(self):
    self.EmitPreprocessor('endif')

  def EmitCode(self, code):
    self.EmitIndented('%s;' % code)

  def EmitFunctionBeginA(self, function_name, params, return_type):
    self.EmitIndented('%s %s(%s) {' %
                      (return_type, function_name,
                       ', '.join(['%s %s' % (t, n) for (t, n) in params])))
    self.PushIndent()

  def EmitFunctionEnd(self):
    self.PopIndent()
    self.EmitIndented('}')

  def EmitAsmBegin(self):
    self.EmitIndented('asm volatile(')
    self.PushIndent()

  def EmitAsmMapping(self, elements):
    if elements:
      self.EmitIndented(': ' + ', '.join(
          ['[%s] "%s"(%s)' % (k, v[1], v[0]) for (k, v) in elements]))
    else:
      self.EmitIndented(':')

  def EmitClobbers(self, elements):
    if elements:
      self.EmitIndented(': ' + ', '.join(['"%s"' % c for c in elements]))
    else:
      self.EmitIndented(':')

  def EmitAsmEnd(self, registers):
    self.EmitAsmMapping(registers.MappedOutputParameters())
    self.EmitAsmMapping(registers.MappedParameters())
    self.EmitClobbers(registers.Clobbers() + ['cc', 'memory'])
    self.PopIndent()
    self.EmitIndented(');')

  def EmitComment(self, comment):
    self.EmitIndented('// ' + comment)

  def EmitNumericalLabel(self, label):
    self.EmitIndented('"%d:"' % label)

  def EmitOp1(self, op, param1):
    self.PushOp(op)
    self.EmitIndented('"%s %s\\n"' % (op, param1))

  def EmitOp2(self, op, param1, param2):
    self.PushOp(op)
    self.EmitIndented('"%s %s, %s\\n"' % (op, param1, param2))

  def EmitOp3(self, op, param1, param2, param3):
    self.PushOp(op)
    self.EmitIndented('"%s %s, %s, %s\\n"' % (op, param1, param2, param3))

  def EmitAdd(self, destination, source, param):
    self.EmitOp3('add', destination, source, param)

  def EmitSubs(self, destination, source, param):
    self.EmitOp3('subs', destination, source, param)

  def EmitSub(self, destination, source, param):
    self.EmitOp3('sub', destination, source, param)

  def EmitMul(self, destination, source, param):
    self.EmitOp3('mul', destination, source, param)

  def EmitMov(self, param1, param2):
    self.EmitOp2('mov', param1, param2)

  def EmitVMovl(self, mov_type, destination, source):
    wide_type = _WideType(mov_type)
    destination = _AppendType(wide_type, destination)
    source = _AppendType(mov_type, _Cast(source.register_bits / 2, source))
    if _UnsignedType(mov_type):
      self.EmitOp2('uxtl', destination, source)
    else:
      self.EmitOp2('sxtl', destination, source)

  def EmitVMovl2(self, mov_type, destination_1, destination_2, source):
    wide_type = _WideType(mov_type)
    if (destination_1.register_bits != source.register_bits or
        destination_2.register_bits != source.register_bits):
      raise ArgumentError('Register sizes do not match.')
    if _UnsignedType(mov_type):
      self.EmitOp2('uxtl2',
                   _AppendType(wide_type, destination_2),
                   _AppendType(mov_type, source))
      self.EmitOp2('uxtl',
                   _AppendType(wide_type, destination_1),
                   _AppendType(mov_type,
                               _Cast(source.register_bits / 2, source)))
    else:
      self.EmitOp2('sxtl2',
                   _AppendType(wide_type, destination_2),
                   _AppendType(mov_type, source))
      self.EmitOp2('sxtl',
                   _AppendType(wide_type, destination_1),
                   _AppendType(mov_type,
                               _Cast(source.register_bits / 2, source)))

  def EmitVMax(self, max_type, destination, source_1, source_2):
    if _UnsignedType(max_type):
      self.EmitOp3('umax',
                   _AppendType(max_type, destination),
                   _AppendType(max_type, source_1),
                   _AppendType(max_type, source_2))
    else:
      self.EmitOp3('smax',
                   _AppendType(max_type, destination),
                   _AppendType(max_type, source_1),
                   _AppendType(max_type, source_2))

  def EmitVMin(self, min_type, destination, source_1, source_2):
    if _UnsignedType(min_type):
      self.EmitOp3('umin',
                   _AppendType(min_type, destination),
                   _AppendType(min_type, source_1),
                   _AppendType(min_type, source_2))
    else:
      self.EmitOp3('smin',
                   _AppendType(min_type, destination),
                   _AppendType(min_type, source_1),
                   _AppendType(min_type, source_2))

  def EmitBeqBack(self, label):
    self.EmitOp1('beq', '%db' % label)

  def EmitBeqFront(self, label):
    self.EmitOp1('beq', '%df' % label)

  def EmitBgtBack(self, label):
    self.EmitOp1('bgt', '%db' % label)

  def EmitBgtFront(self, label):
    self.EmitOp1('bgt', '%df' % label)

  def EmitBleBack(self, label):
    self.EmitOp1('ble', '%db' % label)

  def EmitBleFront(self, label):
    self.EmitOp1('ble', '%df' % label)

  def EmitBneBack(self, label):
    self.EmitOp1('bne', '%db' % label)

  def EmitBneFront(self, label):
    self.EmitOp1('bne', '%df' % label)

  def EmitVAdd(self, add_type, destination, source_1, source_2):
    destination, source_1, source_2 = _MakeCompatibleDown(destination, source_1,
                                                          source_2)
    if _FloatType(add_type):
      self.EmitOp3('fadd',
                   _AppendType(add_type, destination),
                   _AppendType(add_type, source_1),
                   _AppendType(add_type, source_2))
    else:
      self.EmitOp3('add',
                   _AppendType(add_type, destination),
                   _AppendType(add_type, source_1),
                   _AppendType(add_type, source_2))

  def EmitVAddw(self, add_type, destination, source_1, source_2):
    wide_type = _WideType(add_type)
    destination = _AppendType(wide_type, destination)
    source_1 = _AppendType(wide_type, source_1)
    source_2 = _AppendType(add_type, source_2)
    if _UnsignedType(add_type):
      self.EmitOp3('uaddw', destination, source_1, source_2)
    else:
      self.EmitOp3('saddw', destination, source_1, source_2)

  def EmitVSub(self, sub_type, destination, source_1, source_2):
    destination, source_1, source_2 = _MakeCompatibleDown(destination, source_1,
                                                          source_2)
    if _FloatType(sub_type):
      self.EmitOp3('fsub',
                   _AppendType(sub_type, destination),
                   _AppendType(sub_type, source_1),
                   _AppendType(sub_type, source_2))
    else:
      self.EmitOp3('sub',
                   _AppendType(sub_type, destination),
                   _AppendType(sub_type, source_1),
                   _AppendType(sub_type, source_2))

  def EmitVCvt(self, cvt_to, cvt_from, destination, source):
    if cvt_to == 'f32' and cvt_from == 's32':
      self.EmitOp2('scvtf',
                   _AppendType('f32', destination), _AppendType('s32', source))
    elif cvt_to == 'f32' and cvt_from == 'u32':
      self.EmitOp2('ucvtf',
                   _AppendType('f32', destination), _AppendType('u32', source))
    elif cvt_to == 's32' and cvt_from == 'f32':
      self.EmitOp2('fcvtzs',
                   _AppendType('s32', destination), _AppendType('f32', source))
    else:
      raise ArgumentError('Convert not supported, to: %s from: %s' % (cvt_to,
                                                                      cvt_from))

  def EmitVDup(self, dup_type, destination, source):
    if (isinstance(source, _GeneralRegister) or
        isinstance(source, _MappedParameter)):
      self.EmitOp2('dup',
                   _AppendType(dup_type, destination),
                   _Cast(_TypeBits(dup_type), source))
    else:
      self.EmitOp2('dup',
                   _AppendType(dup_type, destination),
                   _AppendType(dup_type, source))

  def EmitVMov(self, mov_type, destination, source):
    if isinstance(source, _ImmediateConstant):
      self.EmitOp2('movi', _AppendType(mov_type, destination), source)
    elif (isinstance(source, _GeneralRegister) or
          isinstance(source, _MappedParameter)):
      self.EmitOp2('mov',
                   _AppendType(mov_type, destination),
                   _Cast(_TypeBits(mov_type), source))
    else:
      self.EmitOp2('mov', _AppendType(8, destination), _AppendType(8, source))

  def EmitVQmovn(self, mov_type, destination, source):
    narrow_type = _NarrowType(mov_type)
    if destination.register_bits * 2 == source.register_bits:
      self.EmitOp2('sqxtn',
                   _AppendType(narrow_type, destination),
                   _AppendType(mov_type, source))
    elif destination.register_bits == source.register_bits:
      self.EmitOp2('sqxtn',
                   _AppendType(narrow_type,
                               _Cast(destination.register_bits / 2,
                                     destination)),
                   _AppendType(mov_type, source))

  def EmitVQmovn2(self, mov_type, destination, source_1, source_2):
    narrow_type = _NarrowType(mov_type)
    if (destination.register_bits != source_1.register_bits or
        destination.register_bits != source_2.register_bits):
      raise ArgumentError('Register sizes do not match.')
    self.EmitOp2('sqxtn',
                 _AppendType(narrow_type,
                             _Cast(destination.register_bits / 2, destination)),
                 _AppendType(mov_type, source_1))
    self.EmitOp2('sqxtn2',
                 _AppendType(narrow_type, destination),
                 _AppendType(mov_type, source_2))

  def EmitVQmovun(self, mov_type, destination, source):
    narrow_type = _NarrowType(mov_type)
    if destination.register_bits * 2 == source.register_bits:
      self.EmitOp2('sqxtun',
                   _AppendType(narrow_type, destination),
                   _AppendType(mov_type, source))
    elif destination.register_bits == source.register_bits:
      self.EmitOp2('sqxtun',
                   _AppendType(narrow_type,
                               _Cast(destination.register_bits / 2,
                                     destination)),
                   _AppendType(mov_type, source))

  def EmitVQmovun2(self, mov_type, destination, source_1, source_2):
    narrow_type = _NarrowType(mov_type)
    if (destination.register_bits != source_1.register_bits or
        destination.register_bits != source_2.register_bits):
      raise ArgumentError('Register sizes do not match.')
    self.EmitOp2('sqxtun',
                 _AppendType(narrow_type,
                             _Cast(destination.register_bits / 2, destination)),
                 _AppendType(mov_type, source_1))
    self.EmitOp2('sqxtun2',
                 _AppendType(narrow_type, destination),
                 _AppendType(mov_type, source_2))

  def EmitVMul(self, mul_type, destination, source_1, source_2):
    destination, source_1, source_2 = _MakeCompatibleDown(destination, source_1,
                                                          source_2)
    if _FloatType(mul_type):
      self.EmitOp3('fmul',
                   _AppendType(mul_type, destination),
                   _AppendType(mul_type, source_1),
                   _AppendType(mul_type, source_2))
    else:
      self.EmitOp3('mul',
                   _AppendType(mul_type, destination),
                   _AppendType(mul_type, source_1),
                   _AppendType(mul_type, source_2))

  def EmitVMulScalar(self, mul_type, destination, source_1, source_2):
    self.EmitOp3('mul',
                 _AppendType(mul_type, destination),
                 _AppendType(mul_type, source_1),
                 _AppendType(mul_type, source_2))

  def EmitVMull(self, mul_type, destination, source_1, source_2):
    wide_type = _WideType(mul_type)
    if _UnsignedType(mul_type):
      self.EmitOp3('umull',
                   _AppendType(wide_type, destination),
                   _AppendType(mul_type, source_1),
                   _AppendType(mul_type, source_2))
    else:
      self.EmitOp3('smull',
                   _AppendType(wide_type, destination),
                   _AppendType(mul_type, source_1),
                   _AppendType(mul_type, source_2))

  def EmitVPadd(self, add_type, destination, source_1, source_2):
    self.EmitOp3('addp',
                 _AppendType(add_type, destination),
                 _AppendType(add_type, source_1),
                 _AppendType(add_type, source_2))

  def EmitVPaddl(self, add_type, destination, source):
    wide_type = _WideType(add_type)
    if _UnsignedType(add_type):
      self.EmitOp2('uaddlp',
                   _AppendType(wide_type, destination),
                   _AppendType(add_type, source))
    else:
      self.EmitOp2('saddlp',
                   _AppendType(wide_type, destination),
                   _AppendType(add_type, source))

  def EmitVPadal(self, add_type, destination, source):
    wide_type = _WideType(add_type)
    if _UnsignedType(add_type):
      self.EmitOp2('uadalp',
                   _AppendType(wide_type, destination),
                   _AppendType(add_type, source))
    else:
      self.EmitOp2('sadalp',
                   _AppendType(wide_type, destination),
                   _AppendType(add_type, source))

  def EmitLdr(self, register, value):
    self.EmitOp2('ldr', _Cast(32, register), _Cast(None, value))

  def EmitVLoad(self, load_no, load_type, destination, source):
    self.EmitVLoadA(load_no, load_type, [destination], source)

  def EmitVLoadA(self, load_no, load_type, destinations, source):
    if source.dereference_increment:
      increment = sum(
          [_LoadStoreSize(destination) for destination in destinations]) / 8
      self.EmitVLoadAPostIncrement(load_no, load_type, destinations, source,
                                   self.ImmediateConstant(increment))
    else:
      self.EmitVLoadAPostIncrement(load_no, load_type, destinations, source,
                                   None)

  def EmitVLoadAPostIncrement(self, load_no, load_type, destinations, source,
                              increment):
    """Generate assembly to load memory to registers and increment source."""
    if len(destinations) == 1 and destinations[0].lane is -1:
      destination = '{%s}' % _AppendType(load_type, destinations[0])
      if increment:
        self.EmitOp3('ld%dr' % load_no, destination, source, increment)
      else:
        self.EmitOp2('ld%dr' % load_no, destination, source)
      return

    destination_list = _RegisterList(load_type, destinations)
    if increment:
      self.EmitOp3('ld%d' % load_no, destination_list, source, increment)
    else:
      self.EmitOp2('ld%d' % load_no, destination_list, source)

  def EmitVLoadAE(self,
                  load_type,
                  elem_count,
                  destinations,
                  source,
                  alignment=None):
    """Generate assembly to load an array of elements of given size."""
    bits_to_load = load_type * elem_count
    min_bits = min([destination.register_bits for destination in destinations])
    max_bits = max([destination.register_bits for destination in destinations])

    if min_bits is not max_bits:
      raise ArgumentError('Cannot mix double and quad loads.')

    if len(destinations) * min_bits < bits_to_load:
      raise ArgumentError('To few destinations: %d to load %d bits.' %
                          (len(destinations), bits_to_load))

    leftover_loaded = 0
    while bits_to_load > 0:
      if bits_to_load >= 4 * min_bits:
        self.EmitVLoadA(1, 32, destinations[:4],
                        self.DereferenceIncrement(source, alignment))
        bits_to_load -= 4 * min_bits
        destinations = destinations[4:]
      elif bits_to_load >= 3 * min_bits:
        self.EmitVLoadA(1, 32, destinations[:3],
                        self.DereferenceIncrement(source, alignment))
        bits_to_load -= 3 * min_bits
        destinations = destinations[3:]
      elif bits_to_load >= 2 * min_bits:
        self.EmitVLoadA(1, 32, destinations[:2],
                        self.DereferenceIncrement(source, alignment))
        bits_to_load -= 2 * min_bits
        destinations = destinations[2:]
      elif bits_to_load >= min_bits:
        self.EmitVLoad(1, 32, destinations[0],
                       self.DereferenceIncrement(source, alignment))
        bits_to_load -= min_bits
        destinations = destinations[1:]
      elif bits_to_load >= 64:
        self.EmitVLoad(1, 32,
                       _Cast(64, destinations[0]),
                       self.DereferenceIncrement(source))
        bits_to_load -= 64
        leftover_loaded += 64
      elif bits_to_load >= 32:
        self.EmitVLoad(1, 32,
                       self.Lane(32, destinations[0], leftover_loaded / 32),
                       self.DereferenceIncrement(source))
        bits_to_load -= 32
        leftover_loaded += 32
      elif bits_to_load >= 16:
        self.EmitVLoad(1, 16,
                       self.Lane(16, destinations[0], leftover_loaded / 16),
                       self.DereferenceIncrement(source))
        bits_to_load -= 16
        leftover_loaded += 16
      elif bits_to_load is 8:
        self.EmitVLoad(1, 8,
                       self.Lane(8, destinations[0], leftover_loaded / 8),
                       self.DereferenceIncrement(source))
        bits_to_load -= 8
        leftover_loaded += 8
      else:
        raise ArgumentError('Wrong leftover: %d' % bits_to_load)

  def EmitVLoadE(self, load_type, count, destination, source, alignment=None):
    self.EmitVLoadAE(load_type, count, [destination], source, alignment)

  def EmitVLoadAllLanes(self, load_no, load_type, destination, source):
    new_destination = destination.Copy()
    new_destination.lane = -1
    new_destination.lane_bits = load_type
    self.EmitVLoad(load_no, load_type, new_destination, source)

  def EmitVLoadOffset(self, load_no, load_type, destination, source, offset):
    self.EmitVLoadOffsetA(load_no, load_type, [destination], source, offset)

  def EmitVLoadOffsetA(self, load_no, load_type, destinations, source, offset):
    assert len(destinations) <= 4
    self.EmitOp3('ld%d' % load_no,
                 _RegisterList(load_type, destinations), source, offset)

  def EmitPld(self, load_address_register):
    self.EmitOp2('prfm', 'pldl1keep', '[%s]' % load_address_register)

  def EmitPldOffset(self, load_address_register, offset):
    self.EmitOp2('prfm', 'pldl1keep',
                 '[%s, %s]' % (load_address_register, offset))

  def EmitVShl(self, shift_type, destination, source, shift):
    self.EmitOp3('sshl',
                 _AppendType(shift_type, destination),
                 _AppendType(shift_type, source), _AppendType('i32', shift))

  def EmitVStore(self, store_no, store_type, source, destination):
    self.EmitVStoreA(store_no, store_type, [source], destination)

  def EmitVStoreA(self, store_no, store_type, sources, destination):
    if destination.dereference_increment:
      increment = sum([_LoadStoreSize(source) for source in sources]) / 8
      self.EmitVStoreAPostIncrement(store_no, store_type, sources, destination,
                                    self.ImmediateConstant(increment))
    else:
      self.EmitVStoreAPostIncrement(store_no, store_type, sources, destination,
                                    None)

  def EmitVStoreAPostIncrement(self, store_no, store_type, sources, destination,
                               increment):
    source_list = _RegisterList(store_type, sources)
    if increment:
      self.EmitOp3('st%d' % store_no, source_list, destination, increment)
    else:
      self.EmitOp2('st%d' % store_no, source_list, destination)

  def EmitVStoreAE(self,
                   store_type,
                   elem_count,
                   sources,
                   destination,
                   alignment=None):
    """Generate assembly to store an array of elements of given size."""
    bits_to_store = store_type * elem_count
    min_bits = min([source.register_bits for source in sources])
    max_bits = max([source.register_bits for source in sources])

    if min_bits is not max_bits:
      raise ArgumentError('Cannot mix double and quad stores.')

    if len(sources) * min_bits < bits_to_store:
      raise ArgumentError('To few destinations: %d to store %d bits.' %
                          (len(sources), bits_to_store))

    leftover_stored = 0
    while bits_to_store > 0:
      if bits_to_store >= 4 * min_bits:
        self.EmitVStoreA(1, 32, sources[:4],
                         self.DereferenceIncrement(destination, alignment))
        bits_to_store -= 4 * min_bits
        sources = sources[4:]
      elif bits_to_store >= 3 * min_bits:
        self.EmitVStoreA(1, 32, sources[:3],
                         self.DereferenceIncrement(destination, alignment))
        bits_to_store -= 3 * min_bits
        sources = sources[3:]
      elif bits_to_store >= 2 * min_bits:
        self.EmitVStoreA(1, 32, sources[:2],
                         self.DereferenceIncrement(destination, alignment))
        bits_to_store -= 2 * min_bits
        sources = sources[2:]
      elif bits_to_store >= min_bits:
        self.EmitVStore(1, 32, sources[0],
                        self.DereferenceIncrement(destination, alignment))
        bits_to_store -= min_bits
        sources = sources[1:]
      elif bits_to_store >= 64:
        self.EmitVStore(1, 32,
                        _Cast(64, sources[0]),
                        self.DereferenceIncrement(destination, alignment))
        bits_to_store -= 64
        leftover_stored += 64
      elif bits_to_store >= 32:
        self.EmitVStore(1, 32,
                        self.Lane(32, sources[0], leftover_stored / 32),
                        self.DereferenceIncrement(destination))
        bits_to_store -= 32
        leftover_stored += 32
      elif bits_to_store >= 16:
        self.EmitVStore(1, 16,
                        self.Lane(16, sources[0], leftover_stored / 16),
                        self.DereferenceIncrement(destination))
        bits_to_store -= 16
        leftover_stored += 16
      elif bits_to_store >= 8:
        self.EmitVStore(1, 8,
                        self.Lane(8, sources[0], leftover_stored / 8),
                        self.DereferenceIncrement(destination))
        bits_to_store -= 8
        leftover_stored += 8
      else:
        raise ArgumentError('Wrong leftover: %d' % bits_to_store)

  def EmitVStoreE(self, store_type, count, source, destination, alignment=None):
    self.EmitVStoreAE(store_type, count, [source], destination, alignment)

  def EmitVStoreOffset(self, store_no, store_type, source, destination, offset):
    self.EmitVStoreOffsetA(store_no, store_type, [source], destination, offset)

  def EmitVStoreOffsetA(self, store_no, store_type, sources, destination,
                        offset):
    self.EmitOp3('st%d' % store_no,
                 _RegisterList(store_type, sources), destination, offset)

  def EmitVStoreOffsetE(self, store_type, count, source, destination, offset):
    if store_type is not 32:
      raise ArgumentError('Unsupported store_type: %d' % store_type)

    if count == 1:
      self.EmitVStoreOffset(1, 32,
                            self.Lane(32, source, 0),
                            self.Dereference(destination, None), offset)
    elif count == 2:
      self.EmitVStoreOffset(1, 32,
                            _Cast(64, source),
                            self.Dereference(destination, None), offset)
    elif count == 3:
      self.EmitVStore(1, 32,
                      _Cast(64, source),
                      self.DereferenceIncrement(destination, None))
      self.EmitVStoreOffset(1, 32,
                            self.Lane(32, source, 2),
                            self.Dereference(destination, None), offset)
      self.EmitSub(destination, destination, self.ImmediateConstant(8))
    elif count == 4:
      self.EmitVStoreOffset(1, 32, source,
                            self.Dereference(destination, None), offset)
    else:
      raise ArgumentError('To many elements: %d' % count)

  def EmitVSumReduce(self, reduce_type, elem_count, reduce_count, destinations,
                     sources):
    """Generate assembly to perform n-fold horizontal sum reduction."""
    if reduce_type is not 'u32':
      raise ArgumentError('Unsupported reduce: %s' % reduce_type)

    if (elem_count + 3) / 4 > len(destinations):
      raise ArgumentError('To few destinations: %d (%d needed)' %
                          (len(destinations), (elem_count + 3) / 4))

    if elem_count * reduce_count > len(sources) * 4:
      raise ArgumentError('To few sources: %d' % len(sources))

    if reduce_count <= 1:
      raise ArgumentError('Unsupported reduce_count: %d' % reduce_count)

    sources = [_Cast(128, source) for source in sources]
    destinations = [_Cast(128, destination) for destination in destinations]

    while reduce_count > 1:
      if len(sources) % 2 == 1:
        sources.append(sources[-1])

      if reduce_count == 2:
        for i in range(len(destinations)):
          self.EmitVPadd(reduce_type, destinations[i], sources[2 * i],
                         sources[2 * i + 1])
        return
      else:
        sources_2 = []
        for i in range(len(sources) / 2):
          self.EmitVPadd(reduce_type, sources[2 * i], sources[2 * i],
                         sources[2 * i + 1])
          sources_2.append(sources[2 * i])
        reduce_count /= 2
        sources = sources_2

  def EmitVUzp1(self, uzp_type, destination, source_1, source_2):
    self.EmitOp3('uzp1',
                 _AppendType(uzp_type, destination),
                 _AppendType(uzp_type, source_1),
                 _AppendType(uzp_type, source_2))

  def EmitVUzp2(self, uzp_type, destination, source_1, source_2):
    self.EmitOp3('uzp2',
                 _AppendType(uzp_type, destination),
                 _AppendType(uzp_type, source_1),
                 _AppendType(uzp_type, source_2))

  def EmitVUzp(self, uzp_type, destination_1, destination_2, source_1,
               source_2):
    self.EmitVUzp1(uzp_type, destination_1, source_1, source_2)
    self.EmitVUzp2(uzp_type, destination_2, source_1, source_2)

  def EmitVTrn1(self, trn_type, destination, source_1, source_2):
    self.EmitOp3('trn1',
                 _AppendType(trn_type, destination),
                 _AppendType(trn_type, source_1),
                 _AppendType(trn_type, source_2))

  def EmitVTrn2(self, trn_type, destination, source_1, source_2):
    self.EmitOp3('trn2',
                 _AppendType(trn_type, destination),
                 _AppendType(trn_type, source_1),
                 _AppendType(trn_type, source_2))

  def EmitVTrn(self, trn_type, destination_1, destination_2, source_1,
               source_2):
    self.EmitVTrn1(trn_type, destination_1, source_1, source_2)
    self.EmitVTrn2(trn_type, destination_2, source_1, source_2)

  def EmitColBlockStride(self, cols, stride, new_stride):
    assert cols in [1, 2, 3, 4, 5, 6, 7, 8]
    if cols in [5, 6, 7]:
      self.EmitSub(new_stride, stride, self.ImmediateConstant(4))

  def EmitLoadColBlock(self, registers, load_type, cols, elements, block,
                       input_address, stride):
    assert cols is len(block)
    assert load_type is 8

    input_deref = self.Dereference(input_address, None)
    input_deref_increment = self.DereferenceIncrement(input_address, None)

    if cols is 1:
      for i in range(elements):
        self.EmitVLoadOffset(1, 8,
                             self.Lane(8, block[0], i), input_deref, stride)
      self.EmitPld(input_address)
      return block
    elif cols is 2:
      temp = [registers.DoubleRegister() for unused_i in range(2)]
      for i in range(elements):
        self.EmitVLoadOffset(1, 16,
                             self.Lane(16, block[i / 4], i % 4), input_deref,
                             stride)
      self.EmitPld(input_address)
      self.EmitVUzp(8, temp[0], temp[1], block[0], block[1])
      registers.FreeRegisters(block)
      return temp
    elif cols is 3:
      for i in range(elements):
        self.EmitVLoadOffsetA(3, 8, [self.Lane(8, row, i) for row in block],
                              input_deref, stride)
      self.EmitPld(input_address)
      return block
    elif cols is 4:
      temp = [registers.DoubleRegister() for unused_i in range(4)]
      for i in range(elements):
        self.EmitVLoadOffset(1, 32,
                             self.Lane(32, block[i % 4], i / 4), input_deref,
                             stride)
      self.EmitPld(input_address)
      self.EmitVTrn(16, temp[0], temp[2], block[0], block[2])
      self.EmitVTrn(16, temp[1], temp[3], block[1], block[3])
      self.EmitVTrn(8, block[0], block[1], temp[0], temp[1])
      self.EmitVTrn(8, block[2], block[3], temp[2], temp[3])
      registers.FreeRegisters(temp)
      return block
    elif cols is 5:
      temp = [registers.DoubleRegister() for unused_i in range(4)]
      for i in range(elements):
        self.EmitVLoad(1, 32,
                       self.Lane(32, block[i % 4], i / 4),
                       input_deref_increment)
        self.EmitVLoadOffset(1, 8,
                             self.Lane(8, block[4], i), input_deref, stride)
      self.EmitPld(input_address)
      self.EmitVTrn(16, temp[0], temp[2], block[0], block[2])
      self.EmitVTrn(16, temp[1], temp[3], block[1], block[3])
      self.EmitVTrn(8, block[0], block[1], temp[0], temp[1])
      self.EmitVTrn(8, block[2], block[3], temp[2], temp[3])
      registers.FreeRegisters(temp)
      return block
    elif cols is 6:
      temp = [registers.DoubleRegister() for unused_i in range(6)]
      for i in range(elements):
        self.EmitVLoad(1, 32,
                       self.Lane(32, block[i % 4], i / 4),
                       input_deref_increment)
        self.EmitVLoadOffset(1, 16,
                             self.Lane(16, block[4 + i / 4], i % 4),
                             input_deref, stride)
      self.EmitPld(input_address)
      self.EmitVTrn(16, temp[0], temp[2], block[0], block[2])
      self.EmitVTrn(16, temp[1], temp[3], block[1], block[3])
      self.EmitVUzp(8, temp[4], temp[5], block[4], block[5])
      self.EmitVTrn(8, block[0], block[1], temp[0], temp[1])
      self.EmitVTrn(8, block[2], block[3], temp[2], temp[3])
      registers.FreeRegisters(
          [block[4], block[5], temp[0], temp[1], temp[2], temp[3]])
      return [block[0], block[1], block[2], block[3], temp[4], temp[5]]
    elif cols is 7:
      temp = [registers.DoubleRegister() for unused_i in range(4)]
      for i in range(elements):
        self.EmitVLoad(1, 32,
                       self.Lane(32, block[i % 4], i / 4),
                       input_deref_increment)
        self.EmitVLoadOffsetA(3, 8,
                              [self.Lane(8, row, i) for row in block[4:]],
                              input_deref, stride)
      self.EmitPld(input_address)
      self.EmitVTrn1(16, temp[0], block[0], block[2])
      self.EmitVTrn2(16, temp[2], block[0], block[2])
      self.EmitVTrn1(16, temp[1], block[1], block[3])
      self.EmitVTrn2(16, temp[3], block[1], block[3])
      self.EmitVTrn1(8, block[0], temp[0], temp[1])
      self.EmitVTrn2(8, block[1], temp[0], temp[1])
      self.EmitVTrn1(8, block[2], temp[2], temp[3])
      self.EmitVTrn2(8, block[3], temp[2], temp[3])
      registers.FreeRegisters(temp)
      return block
    elif cols is 8:
      temp = [registers.DoubleRegister() for unused_i in range(8)]
      for i in range(elements):
        self.EmitVLoadOffset(1, 32, block[i], input_deref, stride)
      self.EmitPld(input_address)
      self.EmitVTrn(8, temp[0], temp[1], block[0], block[1])
      self.EmitVTrn(8, temp[2], temp[3], block[2], block[3])
      self.EmitVTrn(8, temp[4], temp[5], block[4], block[5])
      self.EmitVTrn(8, temp[6], temp[7], block[6], block[7])
      self.EmitVTrn(16, block[0], block[2], temp[0], temp[2])
      self.EmitVTrn(16, block[1], block[3], temp[1], temp[3])
      self.EmitVTrn(16, block[4], block[6], temp[4], temp[6])
      self.EmitVTrn(16, block[5], block[7], temp[5], temp[7])
      self.EmitVTrn(32, temp[0], temp[4], block[0], block[4])
      self.EmitVTrn(32, temp[1], temp[5], block[1], block[5])
      self.EmitVTrn(32, temp[2], temp[6], block[2], block[6])
      self.EmitVTrn(32, temp[3], temp[7], block[3], block[7])
      registers.FreeRegisters(block)
      return temp
    else:
      assert False

  def Dereference(self, value, unused_alignment=None):
    new_value = value.Copy()
    new_value.dereference = True
    return new_value

  def DereferenceIncrement(self, value, alignment=None):
    new_value = self.Dereference(value, alignment).Copy()
    new_value.dereference_increment = True
    return new_value

  def ImmediateConstant(self, value):
    return _ImmediateConstant(value)

  def AllLanes(self, value):
    return '%s[]' % value

  def Lane(self, bits, value, lane):
    new_value = value.Copy()
    if bits * (lane + 1) > new_value.register_bits:
      raise ArgumentError('Lane to big: (%d + 1) x %d > %d' %
                          (lane, bits, new_value.register_bits))
    new_value.lane = lane
    new_value.lane_bits = bits
    return new_value

  def CreateRegisters(self):
    return _NeonRegisters64Bit()
